"""
Provides a simple Molecule class that we can adapt as we need
Most standard functionality should be served by OpenBabel
Uses AtomData to get properties and whatnot
"""
import itertools
import os, numpy as np
import tempfile

from McUtils.Data import AtomData, UnitsData
from McUtils.Coordinerds import CoordinateSet, ZMatrixCoordinates, CartesianCoordinates3D
import McUtils.Numputils as nput
from McUtils.Zachary import Mesh
import McUtils.Plots as plt
from McUtils.ExternalPrograms import RDMolecule

from ..Modes import PrimitiveCoordinatePicker, RedundantCoordinateGenerator

from .MoleculeInterface import *

from .CoordinateSystems import MolecularEmbedding, ModeEmbedding
from .Evaluator import MolecularEvaluator
from .Hamiltonian import MolecularHamiltonian
from .Properties import *

__all__ = [
    "Molecule",
    "MolecoolException"
]

__reload_hook__ = ["..Modes", ".MoleculeInterface", '.CoordinateSystems', '.Hamiltonian', '.Evaluator', '.Properties']

class Molecule(AbstractMolecule):
    """
    General purpose 'Molecule' class where the 'Molecule' need not be a molecule at all
    """

    def __init__(self,
                 atoms,
                 coords,
                 bonds=None,
                 masses=None,
                 name=None,
                 internals=None,
                 rdmol=None,
                 dipole_surface=None,
                 dipole_derivatives=None,
                 potential_surface=None,
                 potential_derivatives=None,
                 normal_modes=None,
                 source_file=None,
                 guess_bonds=True,
                 charge=None,
                 display_mode=None,
                 **metadata
                 ):
        """
        :param atoms: atoms specified by name, either full name or short
        :type atoms: Iterable[str]
        :param coords: coordinates for the molecule, assumed to be in Bohr by default
        :type coords: np.ndarray
        :param bonds: bond specification for the molecule
        :type bonds: Iterable[Iterable[int]] | None
        :param obmol: OpenBabel molecule for doing conversions
        :type obmol:
        :param charge: Net charge on the molecule
        :type charge: int | None
        :param name: Name for the molecule
        :type name: str | None
        :param name: The internal coordinate specification for the molecule
        :type name: np.ndarray[int] | None
        :param dipole_surface: The dipole surface for the system
        :type dipole_surface: DipoleSurface | None
        :param dipole_derivatives: Derivatives of the dipole surface
        :type dipole_derivatives: Iterable[np.ndarray] | None
        :param potential_surface: The potential surface for the system
        :type potential_surface: PotentialSurface | None
        :param potential_derivatives: Derivatives of the potential surface
        :type potential_derivatives: Iterable[np.ndarray] | None
        :param guess_bonds: Whether or not to guess the bonding arrangement when that would be used
        :type guess_bonds: bool
        :param source_file: The data file the molecule was loaded from
        :type source_file: str
        :param kw: Other bound parameters that might be useful
        :type kw:
        """
        # convert "atoms" into list of atom data
        self._ats = [AtomData[atom] if isinstance(atom, (int, np.integer, str)) else atom for atom in atoms]
        self._mass = np.array([a["Mass"] for a in self._ats]) if masses is None else masses
        coords = CoordinateSet(coords, CartesianCoordinates3D)

        internals = self.canonicalize_internals(internals, self.atoms, coords, bonds, masses=self._mass)
        self.embedding = MolecularEmbedding(self.atomic_masses, coords, internals)
        self._mode_embedding = None

        self._name = name

        # properties to be returned

        self._bonds = bonds
        self.guess_bonds = guess_bonds

        self._src = source_file

        self._rdmol = rdmol
        self._dips = DipoleSurfaceManager(self,
                                          surface=dipole_surface,
                                          derivatives=dipole_derivatives
                                          )
        self._pes = PotentialSurfaceManager(self,
                                            surface=potential_surface,
                                            derivatives=potential_derivatives
                                            )

        self._normal_modes = NormalModesManager(self, normal_modes=normal_modes)

        self.evaluator = MolecularEvaluator(self.embedding, self._normal_modes)
        self.hamiltonian = MolecularHamiltonian(self.embedding,
                                                potential_manager=self._pes,
                                                modes_manager=self._normal_modes,
                                                dipole_manager=self._dips,
                                                )

        metadata['charge'] = charge
        self._meta = metadata

        if display_mode is None:
            display_mode = self.default_display_mode
        self.default_display_mode = display_mode

    @classmethod
    def _auto_spec(cls, atoms, coords, bonds, redundant=False, base_coordinates=None,
                   masses=None,
                   nonredundant_coordinates=None,
                   prune_coordinates=True,
                   pruning_options=None,
                   **opts):
        base_coords = base_coordinates
        if bonds is None:
            bonds = RDMolecule.from_coords(
                                           atoms,
                                           coords * UnitsData.convert("BohrRadius", "Angstroms"),
                                           bonds,
                                           guess_bonds=True
                                           ).bonds
        if redundant and nonredundant_coordinates is None:
            nonredundant_coordinates = base_coords
            base_coords = None
        if redundant and nonredundant_coordinates is not None:
            nonredundant_coordinates = PrimitiveCoordinatePicker.prep_unique_coords(nonredundant_coordinates)
            base_coords = nonredundant_coordinates + ([] if base_coords is None else list(base_coords))
        if base_coords is not None:
            base_coords = PrimitiveCoordinatePicker.prep_unique_coords(base_coords)
        specs = PrimitiveCoordinatePicker(
                atoms,
                [b[:2] for b in bonds],
                base_coords=base_coords,
                **opts
            ).coords
        if prune_coordinates:
            if pruning_options is None:
                pruning_options = {}
            expansion = nput.internal_coordinate_tensors(coords, specs, order=1)[1:]
            if masses is None:
                ats = [AtomData[atom] if isinstance(atom, (int, np.integer, str)) else atom for atom in atoms]
                masses = np.array([a["Mass"] for a in ats])
            prune_pos = RedundantCoordinateGenerator.prune_coordinate_specs(
                expansion,
                masses=masses,
                untransformed_coordinates=np.arange(len(base_coords)) if base_coords is not None else None,
                **pruning_options
            )
            specs = [specs[i] for i in prune_pos]
        spec = {'specs':specs}
        if redundant:
            spec['redundant'] = True
            if nonredundant_coordinates is not None:
                spec['untransformed_coordinates'] = np.arange(len(nonredundant_coordinates))
        return spec
    @classmethod
    def canonicalize_internals(cls, spec, atoms, coords, bonds, relocalize=True, masses=None):
        if isinstance(spec, str) and spec.lower() == 'auto':
            spec = {
                'primitives': 'auto'
            }

        if isinstance(spec, str):
            # if spec.lower() == 'auto':
            #     spec = cls._auto_spec(atoms, coords, bonds)
            # else:
            raise ValueError(f"can't understand internal spec '{spec}'")
        elif isinstance(spec, dict):
            prims = spec.get('primitives')
            if prims is not None:
                spec = spec.copy()
                spec['specs'] = prims
                spec['redundant'] = True
                del spec['primitives']
            subspec = spec.get('specs', '')
            if isinstance(subspec, str):
                if subspec.lower() == 'auto':
                    opts = spec.copy()
                    del opts['specs']
                    if 'relocalize' in opts:
                        relocalize = spec.get('relocalize', relocalize)
                        del opts['relocalize']
                    spec = cls._auto_spec(atoms, coords, bonds, masses=masses, **opts)
                else:
                    raise ValueError(f"can't understand internal spec '{spec}'")
            if spec.get('redundant'):
                spec['relocalize'] = spec.get('relocalize', relocalize)
        return spec

    #region Base Coords
    @property
    def coords(self):
        return self.embedding.coords
    @coords.setter
    def coords(self, coords):
        self.embedding.coords = coords
    @property
    def masses(self):
        return self._mass
    @masses.setter
    def masses(self, masses):
        self._mass = masses
        self.embedding.masses = self.atomic_masses
    @property
    def internals(self):
        return self.embedding.internals
    @property
    def charge(self):
        return self._meta.get('charge', 0)
    @charge.setter
    def charge(self, c):
        self._meta['charge'] = c
    @internals.setter
    def internals(self, spec):
        self.embedding = MolecularEmbedding(
            self.masses,
            self.coords,
            spec
        )
    @property
    def internal_coordinates(self):
        return self.embedding.internal_coordinates
    @property
    def redundant_internal_transformation(self):
        return self.embedding.redundant_internal_transformation

    @property
    def mode_embedding(self):
        if self._mode_embedding is None:
            self._mode_embedding = ModeEmbedding(self.embedding, self.normal_modes)
        return self._mode_embedding
    def get_internals(self, strip_embedding=True):
        return self.embedding.get_internals(strip_embedding=strip_embedding)

    def get_cartesians_by_internals(self, order=None, strip_embedding=False, **kw):
        return self.embedding.get_cartesians_by_internals(order=order, strip_embedding=strip_embedding, **kw)

    def get_internals_by_cartesians(self, order=None, strip_embedding=False, **kw):
        return self.embedding.get_internals_by_cartesians(order=order, strip_embedding=strip_embedding, **kw)

    def get_cartesians_by_modes(self, order=None, **kw):
        return self.mode_embedding.get_cartesians_by_internals(order=order, **kw)

    def get_modes_by_cartesians(self, order=None, **kw):
        return self.mode_embedding.get_internals_by_cartesians(order=order, **kw)

    #endregion

    #region Properties
    @property
    def dipole_surface(self):
        """
        :return:
        :rtype: DipoleSurfaceManager
        """
        return self._dips
    @dipole_surface.setter
    def dipole_surface(self, val):
        if not isinstance(val, DipoleSurfaceManager):
            raise TypeError("`dipole_surface` must be {}".format(
                DipoleSurfaceManager.__name__
            ))
        self._dips = val
    @property
    def dipole_derivatives(self):
        return self.dipole_surface.derivatives
    @dipole_derivatives.setter
    def dipole_derivatives(self, derivs):
        self.dipole_surface.derivatives = derivs
    @property
    def potential_surface(self):
        """
        :return:
        :rtype: PotentialSurfaceManager
        """
        return self._pes
    @potential_surface.setter
    def potential_surface(self, val):
        if not isinstance(val, PotentialSurfaceManager):
            raise TypeError("`potential_surface` must be {}".format(
                PotentialSurfaceManager.__name__
            ))
        self._pes = val
    @property
    def potential_derivatives(self):
        base_derivs = [
            (
                (0 if nput.is_numeric(p) else np.asanyarray(p))
                if p is not None else 0
            )
            for p in self.potential_surface.derivatives
        ]
        n = len(base_derivs)
        for i in range(n): # remove 0 padding
            if isinstance(base_derivs[n - (i+1)], np.ndarray):
                n = n - i
                break
        else:
            n = 0
        return base_derivs[:n]
    @potential_derivatives.setter
    def potential_derivatives(self, derivs):
        self.potential_surface.derivatives = derivs

    def rdkit_optimize(self, force_field_type='mmff', **etc):

        rdmol = self.rdmol.copy()
        rdmol.optimize_structure(force_field_type=force_field_type, **etc)

        return type(self).from_rdmol(rdmol)

    def get_rdkit_potential_expansion(self, order=2, *, optimize=False, force_field_type='mmff'):
        if optimize:
            self = self.rdkit_optimize(force_field_type=force_field_type)
        rdmol = self.rdmol

        ang2bohr = UnitsData.convert("Angstroms", "BohrRadius")
        kcal_mol = UnitsData.convert("Kilocalories/Mole", "Hartrees")

        if order > 2:
            raise NotImplementedError("haven't bothered to add FD beyond the Hessian")

        expansion = []
        eng = rdmol.calculate_energy(force_field_type=force_field_type) * kcal_mol
        expansion.append(eng)
        if order > 0:
            grad = rdmol.calculate_gradient(force_field_type=force_field_type) * kcal_mol / ang2bohr
            expansion.append(grad)
        if order > 1:
            hess = rdmol.calculate_hessian(force_field_type=force_field_type) * kcal_mol / (ang2bohr**2)
            expansion.append(hess)

        if optimize:
            return self, expansion
        else:
            return expansion


    def get_internal_potential_derivatives(self, order=None, reembed=True, strip_embedding=True, zero_gradient=False):
        derivs = self.potential_derivatives
        if zero_gradient:
            derivs = [0] + list(derivs[1:])
        if order is None:
            order = len(derivs)
        return nput.tensor_reexpand(
            self.get_cartesians_by_internals(order, reembed=reembed, strip_embedding=strip_embedding),
            derivs,
            order
        )

    @property
    def normal_modes(self):
        """
        :return:
        :rtype: NormalModesManager
        """
        return self._normal_modes
    @normal_modes.setter
    def normal_modes(self, val):
        if not isinstance(val, NormalModesManager):
            val = NormalModesManager.from_data(self, val)
            # raise TypeError("`normal_modes` must be {}".format(
            #     NormalModesManager.__name__
            # ))
        self._normal_modes = val
    def get_normal_modes(self, masses=None, **opts):
        from ..Modes import NormalModes
        return NormalModes.from_molecule(self, masses=masses, **opts)
    @property
    def metadata(self):
        return self._meta
    @metadata.setter
    def metadata(self, val):
        if not isinstance(val, dict):
            raise TypeError("metadata must be {}".format(
                dict.__name__
            ))
        self._meta = val
    #endregion

    def __repr__(self):
        return "{cls}('{name}', formula='{formula}', shape={shape}, coord_sys={coord_sys})".format(
            cls=type(self).__name__,
            name=self.name,
            formula=self.formula,
            shape=self.coords.shape,
            coord_sys=self.coords.system.name if isinstance(self.coords, CoordinateSet) else 'undefined'
        )

    @property
    def num_atoms(self):
        return len(self._ats)
    @property
    def atom_positions(self):
        """
        A mapping of atom types to positions

        :param spec:
        :type spec:
        :return:
        :rtype:
        """
        pos_map = {}
        for i,a in enumerate(self._ats):
            if a["Symbol"] in pos_map:
                pos_map[a["Symbol"]].append(i)
            else:
                pos_map[a["Symbol"]] = [i]
        return pos_map
    @property
    def dummy_positions(self):
        ats = self.atom_positions
        return ats['X'] if 'X' in ats else []
    @property
    def atoms(self):
        return tuple(a["Symbol"] for a in self._ats)
    # @property
    # def masses(self):
    #     if self._mass is None:
    #         return np.array([a["Mass"] for a in self._ats])
    #     else:
    #         return self._mass
    def _atomic_masses(self):
        m = self.masses
        if min(m) < 100:
            m = m*UnitsData.convert("AtomicMassUnits", "AtomicUnitOfMass")
        return m
    @property
    def atomic_masses(self):
        return self._atomic_masses()
    @property
    def bonds(self):
        if self._bonds is None and self.guess_bonds:
            self._bonds = RDMolecule.from_coords(
                                           self.atoms,
                                           self.coords * UnitsData.convert("BohrRadius", "Angstroms"),
                                           None,
                                           guess_bonds=True
                                           ).bonds
            # self._bonds = self.prop("guessed_bonds", tol=1.05, guess_type=True)
        return self._bonds
    @bonds.setter
    def bonds(self, b):
        self._bonds = b
    @property
    def formula(self):
        return self.prop('chemical_formula')
    @property
    def multiconfig(self):
        return self.coords.multiconfig
    @property
    def name(self):
        if self._name is None:
            return "Unnamed"
        else:
            return self._name
    @property
    def source_file(self):
        return self._src
    @source_file.setter
    def source_file(self, src):
        self._src = src

    @property
    def shape(self):
        return self.coords.shape
    def __len__(self):
        if self.multiconfig:
            return self.coords.shape[0]
        else:
            return 1

    def copy(self):
        import copy
        # mostly just use the default and don't be fancy
        new = copy.copy(self)
        # but we also need to do some stuff where we store objects that
        # reference the molecule
        new.normal_modes = new.normal_modes.copy()
        new.normal_modes.set_molecule(new)
        new.potential_surface = new.potential_surface.copy()
        new.potential_surface.set_molecule(new)
        new.dipole_surface = new.dipole_surface.copy()
        new.dipole_surface.set_molecule(new)
        # new._rdmol = new.rdmol.copy()
        # new.rdmol.set_molecule(new)
        return new

    def take_submolecule(self, pos):
        ats = self.atoms
        atoms = [ats[i] for i in pos]
        masses = self.masses[pos,]
        coords = self.coords[..., pos, :]
        pos = set(pos)
        if self.bonds is not None:
            bonds = [
                b for b in self.bonds
                if b[0] in pos and b[1] in pos
            ]
        else:
            bonds = None
        return type(self)(
            atoms,
            coords,
            masses=masses,
            bonds=bonds
        )

    def prop(self, name, *args, **kwargs):
        from .Properties import MolecularProperties, MolecularPropertyError
        if hasattr(MolecularProperties, name):
            return getattr(MolecularProperties, name)(self, *args, **kwargs)
        else:
            raise MolecularPropertyError("{}.{}: property '{}' unknown".format(
                type(self).__name__,
                'prop',
                name
            ))

    @property
    def fragment_indices(self):
        return MolecularProperties.fragment_indices(self)

    @property
    def fragments(self):
        return MolecularProperties.fragments(self)
    #region Coordinate Embeddings

    @property
    def mass_weighted_coords(self):
        """
        :return:
        :rtype: CoordinateSet
        """
        return MolecularProperties.mass_weighted_coords(self)
        # return self.prop('mass_weighted_coords')
    @property
    def center_of_mass(self):
        """
        :return:
        :rtype: CoordinateSet
        """
        return self.prop('center_of_mass')
    @property
    def inertia_tensor(self):
        """
        :return:
        :rtype: (np.ndarray, np.ndarray)
        """
        return self.prop('inertia_tensor')
    @property
    def inertial_eigensystem(self):
        """
        :return:
        :rtype: (np.ndarray, np.ndarray)
        """
        return self.prop('moments_of_inertia')
    @property
    def moments_of_inertia(self):
        """
        :return:
        :rtype: np.ndarray
        """
        return self.prop('moments_of_inertia')[0]
    @property
    def inertial_axes(self):
        """
        :return:
        :rtype: np.ndarray
        """
        return self.prop('moments_of_inertia')[1]

    @property
    def translation_rotation_modes(self):
        """
        :return:
        :rtype: np.ndarray
        """
        return self.prop('translation_rotation_eigenvectors')

    def get_translation_rotation_projector(self, mass_weighted=False):
        return nput.translation_rotation_projector(
            self.coords,
            self.atomic_masses,
            mass_weighted=mass_weighted,
            return_modes=False
        )
        # L_tr = self.translation_rotation_modes[1]
        # A = np.eye(L_tr.shape[0]) - (L_tr @ L_tr.T)
        # if not mass_weighted:
        #     M = np.diag(np.repeat(1 / np.sqrt(self.masses), 3))
        #     A = M @ A @ M
        # return A

    def get_translation_rotation_invariant_transformation(self,
                                           mass_weighted=False,
                                           strip_embedding=True
                                           ):
        return nput.translation_rotation_invariant_transformation(
            self.coords,
            self.atomic_masses,
            mass_weighted=mass_weighted,
            strip_embedding=strip_embedding
        )


    # @property
    # def zmatrix(self):
    #     """
    #     :return:
    #     :rtype:
    #     """
    #     return self._zmat
    # @zmatrix.setter
    # def zmatrix(self, zmatrix):
    #     """
    #     :return:
    #     :rtype:
    #     """
    #     #TODO: add some validation
    #     zmatrix = np.asanyarray(zmatrix).astype(int)
    #     if zmatrix.shape[1] != 4:
    #         raise ValueError("can't understand Z-matrix {}".format(zmatrix))
    #     self._zmat = zmatrix

    #region Evaluation
    def evaluate(self,
                 func,
                 use_internals=None,
                 deriv_order=None,
                 strip_embedding=False
                 ):
        return self.evaluator.evaluate(
            func,
            use_internals=use_internals,
            deriv_order=deriv_order,
            strip_embedding=strip_embedding
        )
    def evaluate_at(self,
                    func,
                    coords,
                    use_internals=None,
                    deriv_order=None,
                    strip_embedding=False
                    ):
        return self.evaluator.evaluate_at(
            func,
            coords,
            use_internals=use_internals,
            deriv_order=deriv_order,
            strip_embedding=strip_embedding
        )

    def get_displaced_coordinates(self, displacements,
                                  which=None, sel=None, axes=None,
                                  use_internals=False,
                                  coordinate_expansion=None,
                                  strip_embedding=False,
                                  shift=True
                                  ):
        return self.evaluator.get_displaced_coordinates(
            displacements,
            which=which, sel=sel, axes=axes,
            use_internals=use_internals,
            strip_embedding=strip_embedding,
            coordinate_expansion=coordinate_expansion,
            shift=shift
        )

    def get_scan_coordinates(self,
                             domains,
                             internals=False,
                             modes=None,
                             order=None,
                             which=None, sel=None, axes=None,
                             shift=True,
                             coordinate_expansion=None,
                             strip_embedding=False
                             ):
        from ..Modes import NormalModes

        if modes is not None:
            if modes is True:
                modes = self.normal_modes.modes.basis.to_new_modes()
            modes = NormalModes.prep_modes(modes)
            # if order is None:
            #     base_expansion = [modes.coords_by_modes]
            # else:
            #     order = 2
            base_expansion = ModeEmbedding(self.embedding, modes).get_cartesians_by_internals(order=order)
            if coordinate_expansion is not None:
                coordinate_expansion = nput.tensor_reexpand(coordinate_expansion, base_expansion)
            else:
                coordinate_expansion = base_expansion
        return self.evaluator.get_scan_coordinates(
            domains,
            internals=internals,
            which=which, sel=sel, axes=axes,
            shift=shift, coordinate_expansion=coordinate_expansion,
            strip_embedding=strip_embedding
        )

    def get_nearest_displacement_atoms(self,
                                       points,
                                       sel=None, axes=None, weighting_function=None,
                                       return_distances=False
                                       ):
        return self.evaluator.get_nearest_displacement_atoms(
            points,
            sel=sel, axes=axes, weighting_function=weighting_function,
            return_distances=return_distances
        )
    def get_nearest_displacement_coordinates(self,
                                             points,
                                             sel=None, axes=None, weighting_function=None,
                                             modes_nearest=False,
                                             return_distances=False
                                             ):
        return self.evaluator.get_nearest_displacement_coordinates(
            points,
            sel=sel, axes=axes, weighting_function=weighting_function,
            modes_nearest=modes_nearest,
            return_distances=return_distances
        )
    def get_nearest_scan_coordinates(self, domains, sel=None, axes=None):
        return self.evaluator.get_nearest_scan_coordinates(domains, sel=sel, axes=axes)

    def plot_molecule_function(self,
                               function,
                               *,
                               axes,
                               sel=None,
                               embed=False,
                               modes_nearest=False,
                               domain=None,
                               domain_padding=1,
                               plot_points=500,
                               weighting_function=None,
                               mask_function=None,
                               mask_value=0,
                               plot_atoms=False,
                               atom_colors=None,
                               atom_radii=None,
                               plotter=None,
                               epilog=None,
                               **plot_options
                               ):

        if self.coords.ndim > 2:
            raise NotImplementedError("function plotting only supported for one structure at a time")

        axes = np.asanyarray(axes)
        if axes.ndim == 0:
            axes = np.array([axes[:]])
        if len(axes) > 2:
            raise ValueError("can only plot up to 2 axes at a time")

        if domain is None:
            domain = Mesh(self.coords[:, axes]).bounding_box
            if domain_padding is not None:
                if isinstance(domain_padding, (int, float, np.integer, np.floating)):
                    domain_padding = [-domain_padding, domain_padding]
                domain_padding = np.asanyarray(domain_padding)
                if domain_padding.ndim == 1:
                    domain_padding = domain_padding[np.newaxis, :]
                domain = domain + domain_padding

        if isinstance(plot_points, (int, np.integer)):
            plot_points = [plot_points] * len(domain)

        grids = []
        for dom, pts in zip(domain, plot_points):
            grids.append(np.linspace(*dom, pts))
        grids = np.array(np.meshgrid(*grids, indexing='xy'))
        grid_points = np.moveaxis(grids, 0, -1).reshape(-1, len(domain))  # vector of points

        eval_points, dists = self.evaluator.get_nearest_displacement_coordinates(
            grid_points,
            axes=axes,
            sel=sel,
            weighting_function=weighting_function,
            modes_nearest=modes_nearest,
            return_distances=True
        )

        if embed:
            eval_points = self.embed_coords(eval_points)

        values = function(eval_points)
        if mask_function is not None:
            mask = mask_function(values, eval_points, dists)
            values[mask] = mask_value

        if plotter is None:
            if len(grids) == 1:
                plotter = plt.Plot
            else:
                plotter = plt.TriContourPlot

        if plot_atoms:
            if epilog is None:
                epilog = []
            ref = self.coords[:, axes]
            atoms = self._ats
            if sel is not None:
                ref = ref[sel, :]
                atoms = [atoms[i] for i in sel]
            if atom_colors is None:
                atom_colors = [None] * len(atoms)
            atom_colors = [at['IconColor'] if c is None else c for c,at in zip(atom_colors, atoms)]
            if atom_radii is None:
                atom_radii = [None] * len(atoms)
            atom_radii = [at['IconRadius'] if c is None else c for c,at in zip(atom_radii, atoms)]
            epilog = list(epilog) + [
                plt.Disk(
                    crd,
                    r,
                    color=c
                )
                for crd,r,c in zip(ref, atom_radii, atom_colors)
            ]

        if values.ndim > 1 and isinstance(plotter, type) and issubclass(plotter, (plt.Plot, plt.Plot2D)):
            values = np.moveaxis(values, 0, -1)[..., np.newaxis]
            def plotter(subvals,
                        _baseclass=plotter,
                        _subgrids=tuple(np.moveaxis(grid_points, -1, 0)),
                        method=None,
                        **opts
                        ):
                return _baseclass(
                    *_subgrids,
                    subvals.reshape(-1),
                    **opts
                )
            return plt.TensorPlot(values, plot_class=plotter, epilog=epilog, **plot_options)
        else:
            return plotter(*np.moveaxis(grid_points, -1, 0), values, epilog=epilog, **plot_options)

    def get_model(self, potential_specs, dipole=None):
        from ..AnalyticModels import MolecularModel

        if self.internals is None:
            raise ValueError("need internal coordinates to generate analytic model")

        internals = self.internals
        if internals['zmatrix'] is None or internals['conversion'] is not None:
            raise ValueError("only plain Z-matrix coordinate specs currently supported")

        eq_vals = self.internal_coordinates
        embedding_coords = [0, 1, 2, 4, 5, 8]
        good_coords = np.setdiff1d(np.arange(3 * len(self.atoms)), embedding_coords)
        eq_vals = eq_vals.flatten()[good_coords]

        zmat = np.array(internals['zmatrix'])
        def canonicalize_coord(coord_index):
            if coord_index == 0:
                atoms = zmat[1, :2]
            elif coord_index == 1:
                atoms = zmat[2, :2]
            elif coord_index == 2:
                atoms = zmat[2, :3]
            else:
                row = 3 + (coord_index - 3) // 3
                num = 2 + (coord_index - 3) % 3
                atoms = zmat[row, :num]

            if len(atoms) == 2:
                coord_type = 'r'
            elif len(atoms) == 3:
                coord_type = 'a'
            elif len(atoms) == 4:
                coord_type = 't'
            else:
                raise ValueError("bad coord type")
            coord = getattr(MolecularModel, coord_type)(*atoms)

            return coord, atoms

        def canonicalize_spec(coord_index, spec_dict):
            if 'function_type' not in spec_dict and len(spec_dict) == 1:
                function_type, spec_dict = next(iter(spec_dict.items()))
            else:
                function_type = spec_dict['function_type']
            atoms = canonicalize_coord(coord_index)[1]

            scaling = spec_dict.get('scaling', 1)
            params = spec_dict.copy()
            if 'scaling' in params:
                del params['scaling']
            if 'function_type' in params:
                del params['function_type']

            if 'eq' not in params:
                params['eq'] = eq_vals[coord_index]

            return scaling * getattr(MolecularModel, function_type)(*atoms, **params)

        coord_indices = set()

        if isinstance(potential_specs, dict):
            potential_contribs = []
            for idx, spec in potential_specs.items():
                if isinstance(idx, int):
                    idx = [idx]
                    spec = [spec]
                coord_indices.update(idx)
                fns = [canonicalize_spec(i, s) for i,s in zip(idx, spec)]
                f = fns[0]
                for fn in fns[1:]:
                    f = f * fn
                potential_contribs.append(f)
            pot = sum(potential_contribs)
        else:
            pot = potential_specs

        if dipole is not None:
            if len(dipole) != 3:
                raise ValueError("need xyz for dipole contribs")

            dip = []
            for d in dipole:
                if not isinstance(d, dict):
                    dip.append(d)
                    continue

                contribs = []
                for idx, spec in d.items():
                    if isinstance(idx, int):
                        idx = [idx]
                        spec = [spec]
                    coord_indices.update(idx)
                    fns = [canonicalize_spec(i, s) for i, s in zip(idx, spec)]
                    f = fns[0]
                    for fn in fns[1:]:
                        f = f * fn
                    contribs.append(f)
                dip.append(sum(contribs))
        else:
            dip = None

        atoms = set()
        coords = []
        vals = {}
        for idx in sorted(list(coord_indices)):
            c, a = canonicalize_coord(idx)
            atoms.update(a)
            coords.append(c)
            vals[c] = eq_vals[idx]
        masses = self._atomic_masses()

        vals.update({
                MolecularModel.m(i): masses[i]
                for i in sorted(list(atoms))
            })

        return MolecularModel(
            self,
            coords,
            pot,
            dipole=dip,
            values=vals
        )

    def setup_AIMD(self,
                   potential_function,
                   timestep=.5,
                   seed=None,
                   total_energy=None,
                   trajectories=1,
                   sampled_modes=None,
                   initial_energies=None,
                   initial_displacements=None,
                   displaced_coords=None,
                   track_kinetic_energy=False,
                   track_velocities=False
                   ):
        from ..AIMD import AIMDSimulator

        if initial_displacements is not None:
            init_pos = self.get_displaced_coordinates(
                initial_displacements,
                which=displaced_coords,
                internals='reembed'
            )
            sim = AIMDSimulator(
                self.masses,
                init_pos,
                lambda c: -potential_function(c, deriv_order=1)[1].reshape(c.shape),
                timestep=timestep,
                track_kinetic_energy=track_kinetic_energy,
                track_velocities=track_velocities
            )
        else:
            self.potential_derivatives = potential_function(self.coords, deriv_order=2)[1:]

            if total_energy is not None:
                if seed is not None:
                    np.random.seed(seed)
                freqs = self.normal_modes.modes.freqs
                if sampled_modes is None:
                    sampled_modes = list(range(freqs.shape[0]))
                subdirs = np.random.normal(0, 1, size=(trajectories, len(sampled_modes)))

                dirs = np.zeros((trajectories, freqs.shape[0]))
                dirs[:, sampled_modes] = subdirs
                dirs = dirs / np.linalg.norm(dirs, axis=1)[:, np.newaxis] # random unbiased directions

                dirs = dirs / np.sum(np.abs(dirs), axis=1)[:, np.newaxis] # weights in each dimension
                energies = dirs * freqs[np.newaxis, :]
                initial_energies = total_energy * energies / np.sum(np.abs(energies), axis=1)[:, np.newaxis]

            nms = self.normal_modes.modes.basis
            sim = AIMDSimulator(
                self.atomic_masses,
                [self.coords] * len(initial_energies),
                lambda c: -potential_function(c, deriv_order=1)[1].reshape(c.shape),
                velocities=AIMDSimulator.mode_energies_to_velocities(
                    nms.inverse.T,
                    self.atomic_masses,
                    initial_energies,
                    inverse=nms.matrix.T
                ),
                timestep=timestep,
                track_kinetic_energy=track_kinetic_energy,
                track_velocities=track_velocities
            )

        return sim

    def setup_VPT(self,
                  *,
                  states=2,
                  order=2,
                  use_internals=None,
                  **opts
                  ):
        from ..VPT2 import VPTRunner

        if use_internals or use_internals is None:
            return VPTRunner.construct(self, states, order=order, **opts)
        else:
            return VPTRunner.construct(
                [self.atoms, self.coords],
                states,
                potential_derivatives=self.potential_derivatives,
                modes=self.normal_modes.modes.basis,
                order=order,
                **opts
            )

    def get_gmatrix(self, masses=None, use_internals=None):
        if use_internals is None:
            use_internals = self.internal_coordinates is not None

        if not use_internals:
            if masses is None:
                masses = self._atomic_masses()
            else:
                masses = np.asanyarray(masses)
            mass_spec = np.broadcast_to(masses[:, np.newaxis], (len(masses), 3)).flatten()
            return np.diag(1 / mass_spec)
            # raise ValueError("need internal coordinates to calculate the G-matrix")
        else:
            return self.hamiltonian.gmatrix_expansion(0, masses=masses, modes=None)[0]
    @property
    def g_matrix(self):
        """
        Returns the molecular g-matrix for the system
        :return:
        :rtype:
        """
        return self.get_gmatrix()

    @property
    def coriolis_constants(self):
        """
        Returns the molecular g-matrix for the system
        :return:
        :rtype:
        """
        return self.prop('coriolis_constants')

    def bond_length(self, i, j):
        """
        Returns the bond length of the coordinates

        :param i:
        :type i:
        :param j:
        :type j:
        :return:
        :rtype:
        """
        return nput.pts_norms(self.coords[..., i, :], self.coords[..., j, :])
    def bond_angle(self, i, j, k):
        """
        Returns the bond angle of the specified coordinates

        :param i:
        :type i:
        :param j:
        :type j:
        :return:
        :rtype:
        """
        return nput.pts_angles(self.coords[..., i, :], self.coords[..., j, :], self.coords[..., k, :])[0]
    def dihedral(self, i, j, k, l):
        """
        Returns the dihedral angle of the specified coordinates

        :param i:
        :type i:
        :param j:
        :type j:
        :return:
        :rtype:
        """
        return nput.pts_dihedrals(
            self.coords[..., i, :], self.coords[..., j, :],
            self.coords[..., k, :], self.coords[..., l, :]
        )[0]

    def principle_axis_frame(self, sel=None, inverse=False):
        """
        Gets the principle axis frame(s) for the molecule
        :param mol:
        :type mol:
        :param sel: selection of atoms to use when getting the Eckart frame
        :type sel:
        :param inverse: whether to return the inverse of the rotations or not
        :type inverse: bool
        :return:
        :rtype: MolecularTransformation | List[MolecularTransformation]
        """
        return self.prop('principle_axis_transformation', sel=sel, inverse=inverse)

    @property
    def principle_axis_data(self):
        """
        Gets the principle axis embedded coords and embedding parameters for the molecule

        :return:
        :rtype: MolecularTransformation | List[MolecularTransformation]
        """
        return self.prop('principle_axis_data')

    def eckart_frame(self, mol, sel=None, inverse=False, planar_ref_tolerance=None, proper_rotation=False):
        """
        Gets the Eckart frame(s) for the molecule

        :param mol:
        :type mol:
        :param sel: selection of atoms to use when getting the Eckart frame
        :type sel:
        :param inverse: whether to return the inverse of the rotations or not
        :type inverse: bool
        :return:
        :rtype: MolecularTransformation
        """
        return self.prop('eckart_transformation', mol, sel=sel, inverse=inverse,
                         planar_ref_tolerance=planar_ref_tolerance,
                         proper_rotation=proper_rotation)

    def embed_coords(self, crds, sel=None, in_paf=False, planar_ref_tolerance=None, proper_rotation=False):
        """
        Embeds coords in the Eckart frame using `self` as a reference

        :param crds:
        :type crds:
        :return:
        :rtype:
        """

        return self.prop('eckart_embedded_coords', crds, sel=sel, in_paf=in_paf,
                         planar_ref_tolerance=planar_ref_tolerance,
                         proper_rotation=proper_rotation
                         )
    def get_embedding_data(self, crds, sel=None, in_paf=False, planar_ref_tolerance=None, proper_rotation=False):
        """
        Gets the necessary data to embed crds in the Eckart frame using `self` as a reference

        :param crds:
        :type crds:
        :return:
        :rtype:
        """
        return self.prop('eckart_embedding_data', crds, sel=sel, in_paf=in_paf,
                         planar_ref_tolerance=planar_ref_tolerance,
                         proper_rotation=proper_rotation)
    def get_embedded_molecule(self,
                              ref=None,
                              sel=None, planar_ref_tolerance=None,
                              proper_rotation=False,
                              embed_properties=True,
                              load_properties=True
                              ):
        """
        Returns a Molecule embedded in an Eckart frame if ref is not None, otherwise returns
        a principle-axis embedded Molecule
        :return:
        :rtype: Molecule
        """

        if ref is None:
            frame = self.principle_axis_frame(sel=sel, inverse=False)
        else:
            frame = self.eckart_frame(ref, sel=sel, planar_ref_tolerance=planar_ref_tolerance, inverse=False,
                                      proper_rotation=proper_rotation
                                      )
        # self.normal_modes.modes
        new = frame.apply(self)
        if embed_properties:
            # inv_frame = frame
            # if ref is None:
            #     inv_frame = self.principle_axis_frame(inverse=True)
            # else:
            #     inv_frame = self.eckart_frame(ref, inverse=True)
            if load_properties or new._normal_modes._modes is not None:
                new.normal_modes = new.normal_modes.apply_transformation(frame)
            if load_properties or new._pes._surf is not None or new._pes._derivs is not None:
                new.potential_surface = new.potential_surface.apply_transformation(frame)
            if load_properties or new._dips._surf is not None or new._dips._derivs is not None:
                new.dipole_surface = new.dipole_surface.apply_transformation(frame)
        new.source_file = None # for safety

        return new

    #endregion

    #region Input Formats
    @property
    def rdmol(self):
        from McUtils.ExternalPrograms import RDMolecule

        return RDMolecule.from_mol(self, coord_unit="BohrRadius")

    @classmethod
    def from_zmat(cls, zmat, **opts):
        """Little z-matrix importer

        :param zmat:
        :type zmat: str | tuple
        :return:
        :rtype: Molecule
        """

        if isinstance(zmat, str):
            from McUtils.Parsers import StringParser, ZMatPattern
            zmcs = StringParser(ZMatPattern).parse(zmat)
        else:
            zmcs = zmat
        coords = CoordinateSet([zmcs[1]], ZMatrixCoordinates).convert(CartesianCoordinates3D)
        return cls(zmcs[0], coords, zmatrix=zmat[:, (1, 3, 5)], **opts)
    @classmethod
    def from_pybel(cls, mol, **opts):
        """

        :param mol:
        :type mol: pybel.mol
        :return:
        :rtype:
        """
        from McUtils.ExternalPrograms import OpenBabelInterface

        ob = OpenBabelInterface().openbabel
        bonds = list(ob.OBMolBondIter(mol.OBMol))
        atoms = list(mol.atoms)

        opts = dict({'bonds':bonds, 'obmol':mol}, **opts)
        return cls(atoms, [a.coords for a in atoms], **opts)
    @classmethod
    def _from_log_file(cls, file, num=None, **opts):
        from McUtils.GaussianInterface import GaussianLogReader
        with GaussianLogReader(file) as gr:
            parse = gr.parse('StandardCartesianCoordinates', num=num)
        spec, coords = parse['StandardCartesianCoordinates']
        ang2bohr = UnitsData.convert("Angstroms", "AtomicUnitOfLength")
        return cls(
            [int(a[1]) for a in spec],
            CoordinateSet(ang2bohr*np.array(coords), CartesianCoordinates3D),
            **opts
        )
    @classmethod
    def _from_fchk_file(cls, file, **opts):
        from McUtils.GaussianInterface import GaussianFChkReader
        with GaussianFChkReader(file) as gr:
            parse = gr.parse(
                ['Coordinates', 'AtomicNumbers', 'Integer atomic weights', "Real atomic weights"]
            )
        nums = parse["AtomicNumbers"]
        wts = parse['Integer atomic weights']
        masses = parse["Real atomic weights"]

        # print(nums, wts)
        mol = cls(
            [AtomData[a]["Symbol"] + str(b) for a, b in zip(nums, wts)],
            parse["Coordinates"],
            masses=masses,
            **opts
        )
        return mol
    @classmethod
    def from_rdmol(cls, rdmol, **opts):
        return cls(
            rdmol.atoms,
            rdmol.coords * UnitsData.convert("Angstroms", "BohrRadius"),
            bonds=rdmol.bonds,
            **dict(
                rdmol.meta,
                **opts
            )
        )

    @classmethod
    def _from_smiles(cls, smi, add_implicit_hydrogens=False, **opts):
        from McUtils.ExternalPrograms import RDMolecule

        return cls.from_rdmol(RDMolecule.from_smiles(smi, add_implicit_hydrogens=add_implicit_hydrogens), **opts)
    @classmethod
    def _from_name(cls, name, api_key=None, add_implicit_hydrogens=True, **opts):
        from McUtils.ExternalPrograms import ChemSpiderAPI

        smiles = ChemSpiderAPI(api_key).get_compounds_by_name(name, fields="SMILES")
        if len(smiles) == 1:
            return cls._from_smiles(smiles[0]['smiles'], add_implicit_hydrogens=add_implicit_hydrogens, **opts)
        elif len(smiles) > 0:
            return [
                cls._from_smiles(s['smiles'], add_implicit_hydrogens=add_implicit_hydrogens, **opts)
                for s in smiles
            ]
        else:
            raise ValueError(f"{name} didn't resolve to any known compounds with ChemSpider")
    @classmethod
    def _from_sdf(cls, sdf, **opts):
        from McUtils.ExternalPrograms import RDMolecule

        return cls.from_rdmol(RDMolecule.from_sdf(sdf), **opts)
    @classmethod
    def _from_molblock(cls, sdf, **opts):
        from McUtils.ExternalPrograms import RDMolecule

        return cls.from_rdmol(RDMolecule.from_molblock(sdf), **opts)

    @classmethod
    def _from_xyz(cls, xyz, units=None, **opts):
        from McUtils.Parsers import Number, Word
        if xyz[0].isdigit():
            xyz = "\n".join(xyz.splitlines()[2:]) # should confirm that regular `xyz.split("\n", 2)[-1]` would work
        atoms = Word.findall(xyz)
        coords = np.array(Number.findall(xyz)).astype(float).reshape(-1, 3)
        if units is not None:
            coords *= UnitsData.convert(units, "BohrRadius")
        return cls(atoms, coords, **opts)

        # return cls.from_rdmol(RDMolecule.from_molblock(sdf), **opts)

    @classmethod
    def _from_xyz_file(cls, xyz_file, **opts):
        with open(xyz_file) as xyz:
            return cls._from_xyz(xyz.read(), **opts)

    @classmethod
    def from_name(cls, name, **opts):
        return cls.from_string(name, 'name', **opts)

    @classmethod
    def from_string(cls, string, fmt, **opts):
        format_dispatcher = {
            "smi": cls._from_smiles,
            "name": cls._from_name,
            "mol": cls._from_molblock,
            "sdf": cls._from_sdf,
            "xyz": cls._from_xyz
        }

        if fmt in format_dispatcher:
            return format_dispatcher[fmt](string, **opts)
        else:
            with tempfile.NamedTemporaryFile("w+") as tf:
                tf.write(string)
                new = cls.from_file(tf.name, mode=fmt, **opts)
            new.source_file = None # don't want to keep this lying around...
        return new

    @classmethod
    def from_file(cls, file, mode=None, **opts):
        """In general we'll delegate to pybel except for like Fchk and Log files

        :param file:
        :type file:
        :return:
        :rtype:
        """
        import os

        opts['source_file'] = file
        format_dispatcher = {
            "log": cls._from_log_file,
            "fchk": cls._from_fchk_file,
            "smi": cls._from_smiles,
            "mol":cls._from_molblock,
            "sdf":cls._from_sdf,
            "xyz": cls._from_xyz_file
        }

        if mode == None:
            path, ext = os.path.splitext(file)
            ext = ext.lower()
            mode = ext.strip(".")

        if mode in format_dispatcher:
            loader = format_dispatcher[mode]
            return loader(file, **opts)
        else:
            from McUtils.ExternalPrograms import OpenBabelInterface
            try:
                pybel = OpenBabelInterface().pybel
            except ImportError:
                pybel = None
            if pybel is None:
                raise IOError("{} doesn't support file type {} without OpenBabel installed.".format(cls.__name__, mode))
            else:
                mol = next(pybel.readfile(mode, file))
                return cls.from_pybel(mol)

    @classmethod
    def _infer_spec_format(cls, spec):
        if isinstance(spec, str):
            return 'file'

        fmt = None
        try:
            atoms = spec[0]
            if all(isinstance(a, str) for a in atoms):
                return 'standard'
        except:
            pass

        if fmt is None:
            raise ValueError("don't know how to build a molecule from spec {}".format(spec))

    @classmethod
    def from_spec(cls, spec):
        fmt = cls._infer_spec_format(spec)
        if fmt == 'file':
            return cls.from_file(spec)
        elif fmt == 'standard':
            atoms = spec[0]
            coords = spec[1]
            if len(spec) == 2:
                opts = {}
            elif len(spec) == 3:
                opts = spec[2]
            else:
                raise ValueError("too many arguments in {} to build {}".format(spec, cls.__name__))
            return cls(atoms, coords, **opts)
        elif fmt == 'zmat':
            if isinstance(spec[0], str):
                opts = {}
                zmat = spec
            else:
                if len(spec) == 1:
                    opts = {}
                    zmat = spec[0]
                elif len(spec) == 2:
                    opts = spec[1]
                    zmat = spec[0]
                else:
                    raise ValueError("too many arguments in {} to build {}".format(spec, cls.__name__))
            return cls.from_zmat(zmat, **opts)
        else:
            raise NotImplementedError("don't have {} loading from format {} for spec {}".format(cls.__name__, fmt, spec))
    #endregion

    #region Visualization
    highlight_styles = {
        "glow":"green",
        "color":"white"
    }
    def plot(self,
             *geometries,
             figure=None,
             return_objects=False,
             bond_radius=.1,
             atom_radius_scaling=.25,
             atom_style=None,
             bond_style=None,
             highlight_atoms=None,
             highlight_bonds=None,
             highlight_rings=None,
             highlight_styles=None,
             mode=None,#'quality',
             backend='x3d',
             objects=False,
             graphics_class=None,
             cylinder_class=None,
             sphere_class=None,
             animate=None,
             animation_options=None,
             jsmol_load_script=None,
             units="Angstroms",
             **plot_ops
             ):

        if mode is None:
            mode = 'backend'

        if mode == 'jupyter':
            return self.jupyter_viz()
        elif mode == 'jsmol':
            return self.jsmol_viz(script=jsmol_load_script)

        from McUtils.Plots import Graphics3D, Sphere, Cylinder, Line, Disk

        graphics_keys = Graphics3D.known_keys | Graphics3D.opt_keys | Graphics3D.figure_keys
        graphics_opts = {k:plot_ops[k] for k in plot_ops.keys() & graphics_keys}
        plot_ops = {k:plot_ops[k] for k in plot_ops.keys() - graphics_keys}

        if graphics_class is None:
            graphics_class = Graphics3D
        if cylinder_class is None:
            cylinder_class = Line if mode == 'fast' else Cylinder
        if sphere_class is None:
            sphere_class = Disk if mode == 'fast' else Sphere

        if len(geometries) == 0:
            geometries = self.coords
        elif len(geometries) == 1:
            geometries = CoordinateSet(np.asanyarray(geometries[0]), self.coords.system)
        else:
            geometries = CoordinateSet([np.asanyarray(g) for g in geometries], self.coords.system)

        if units is not None:
            geometries = geometries * UnitsData.convert("BohrRadius", units)

        if geometries.ndim == 2:
            geometries = geometries[np.newaxis]
        if animate is None:
            animate = geometries.shape[0] > 1

        geometries = geometries.convert(CartesianCoordinates3D)

        if figure is None:
            figure = graphics_class(backend=backend, **graphics_opts)

        colors = [ at["IconColor"] for at in self._ats ]
        radii = [ atom_radius_scaling * at["IconRadius"] for at in self._ats ]

        bonds = [None] * len(geometries)
        atoms = [None] * len(geometries)

        if atom_style is None:
            atom_style = {}
        elif not isinstance(atom_style, dict):
            atom_style = {i:a for i,a in enumerate(atom_style)}
        base_atom_style = {}
        _atom_style = {i:{} for i in range(len(self._ats))}
        for k,v in atom_style.items():
            if isinstance(k, str):
                base_atom_style[k] = v
            else:
                _atom_style[k] = v
        for k,v in _atom_style.items():
            _atom_style[k] = dict(base_atom_style, **v)
        atom_style = _atom_style


        if highlight_styles is None:
            highlight_styles = self.highlight_styles

        if highlight_rings is not None:
            if highlight_atoms is None:
                highlight_atoms = []
            if highlight_bonds is None:
                highlight_bonds = []
            highlight_atoms = list(highlight_atoms)
            highlight_bonds = list(highlight_bonds)
            for r in highlight_rings:
                highlight_atoms.extend(r)
                highlight_bonds.extend(zip(
                    r, r[1:] + r[:1]
                ))

        if highlight_atoms is not None:
            for k in highlight_atoms:
                if k not in atom_style: atom_style[k] = {}
                atom_style[k].update(highlight_styles)

        if bond_style is None:
            bond_style = {}
        elif not isinstance(bond_style, dict):
            bond_style = {i:a for i,a in enumerate(bond_style)}
        base_bond_style = {}
        _bond_style = {(i,j):{} for i,j in itertools.combinations(range(len(self._ats)), 2)}
        for k,v in bond_style.items():
            if isinstance(k, str):
                base_bond_style[k] = v
            else:
                _bond_style[k] = v
        for k,v in _bond_style.items():
            _bond_style[k] = dict(base_bond_style, **v)
        bond_style = _bond_style
        if highlight_bonds is None and highlight_atoms is not None:
            highlight_bonds = highlight_atoms
        if highlight_bonds is not None:
            for k in highlight_bonds:
                if not nput.is_numeric(k): k = tuple(k)
                if k not in bond_style: bond_style[k] = {}
                bond_style[k].update(highlight_styles)

        for i, geom in enumerate(geometries):
            bond_list = self.bonds
            if bond_style is not False and bond_list is not None:
                bonds[i] = [None] * len(bond_list)
                for j, b in enumerate(bond_list):
                    atom1 = b[0]
                    atom2 = b[1]
                    # i'm not supporting double or triple bonds for not because they're just too much of a pain...
                    # in Mathematica I have some clever code for finding the midpoint between the vectors to draw to
                    # I'm not gonna do that for now because it's not worth it for the quick-and-dirty stuff I want to do
                    p1 = geom[atom1]
                    p2 = geom[atom2]
                    midpoint = (p2 - p1)/2 + p1
                    c1 = colors[atom1]
                    c2 = colors[atom2]

                    base_bstyle = dict(
                        bond_style.get((atom2, atom1), {}),
                        **bond_style.get((atom1, atom2), {})
                    )
                    b_sty_1 = dict(
                        bond_style.get(atom1, {}),
                        **base_bstyle
                    )
                    if b_sty_1.get('color') is None:
                        b_sty_1['color'] = c1
                    cc1 = cylinder_class(
                        p1,
                        midpoint,
                        bond_radius,
                        **plot_ops,
                        **b_sty_1
                    )

                    b_sty_2 = dict(
                        bond_style.get(atom2, {}),
                        **base_bstyle
                    )
                    if b_sty_2.get('color') is None:
                        b_sty_2['color'] = c2
                    cc2 = cylinder_class(
                        midpoint,
                        p2,
                        bond_radius,
                        **plot_ops,
                        **b_sty_2
                    )
                    if objects:
                        bonds[i][j] = (( cc1, cc2 ))
                    else:
                        cyl_1 = cc1.plot(figure)
                        cyl_2 = cc2.plot(figure)
                        if isinstance(cyl_1, (list, tuple)):
                            cyl_1 = cyl_1[0]
                        if isinstance(cyl_2, (list, tuple)):
                            cyl_2 = cyl_2[0]
                        bonds[i][j] = (( cyl_1, cyl_2 ))

            if atom_style is not False:
                atoms[i] = [None] * len(geom)
                for j, stuff in enumerate(zip(colors, radii, geom)):
                    color, radius, coord = stuff
                    a_sty = atom_style.get(j, {})
                    if a_sty.get('color') is None:
                        a_sty['color'] = color

                    sphere = sphere_class(coord, radius, **plot_ops, **a_sty)
                    if objects:
                        atoms[i][j] = sphere
                    else:
                        plops = sphere.plot(figure)
                        if isinstance(plops, tuple):
                            atoms[i][j] = plops[0]
                        else:
                            atoms[i][j] = plops

        if animate:
            if animation_options is None: animation_options = {}
            figure = figure.animate_frames(
                [
                    a + sum([list(b) for b in bl], [])
                    for a,bl in zip(atoms, bonds)
                ],
                **animation_options
            )

        if return_objects:
            return figure, atoms, bonds
        else:
            return figure

    def get_animation_geoms(self, which, extent=.35, steps=8, strip_embedding=True, units=None,
                            coordinate_expansion=None
                            ):
        if isinstance(which, int):
            if coordinate_expansion is None:
                coordinate_expansion = self.get_cartesians_by_internals(2, strip_embedding=strip_embedding)
        else:
            if isinstance(which, np.ndarray):
                if which.ndim == 1:
                    which = which[np.newaxis]
                which = [which]
            coordinate_expansion = which
            which = 0
        geoms = self.get_scan_coordinates(
            [[-extent, extent, steps]],
            which=[which],
            coordinate_expansion=coordinate_expansion
        )
        geoms = np.concatenate([geoms, np.flip(geoms, axis=0)], axis=0)
        if units is not None:
            geoms = geoms * UnitsData.convert("BohrRadius", units)
        return geoms
    def animate_coordinate(self, which, extent=.5, steps=8, return_objects=False, strip_embedding=True,
                           units="Angstroms",
                           backend='jsmol',
                           mode=None,
                           jsmol_load_script=None,
                           coordinate_expansion=None,
                           **plot_opts
                           ):
        if mode is None:
            mode = backend
        if mode == 'jsmol':
            disps = self.format_animation_file(which, format="jmol",
                                               extent=extent, steps=steps, strip_embedding=strip_embedding,
                                               units="Angstroms" if units is None else units,
                                               coordinate_expansion=coordinate_expansion
                                               )
            return self.jsmol_viz(disps, vibrate=True, script=jsmol_load_script)
        else:
            geoms = self.get_animation_geoms(which, extent=extent, steps=steps, strip_embedding=strip_embedding, units=units,
                                             coordinate_expansion=coordinate_expansion)
            return self.plot(geoms, return_objects=return_objects, units=None, backend=backend, **plot_opts)

    def animate_mode(self,
                     which,
                     extent=.5,
                     steps=8,
                     modes=None,
                     coordinate_expansion=None,
                     order=None,
                     normalize=True,
                     mass_weight=False,
                     mass_scale=True,
                     frequency_scale=True,
                     **opts
                     ):
        from ..Modes import NormalModes

        if modes is None:
            modes = self.normal_modes.modes.basis.to_new_modes()
        modes = NormalModes.prep_modes(modes)
        base_expansion = [modes.coords_by_modes]
        if normalize:
            base_expansion = [nput.vec_normalize(base_expansion[0])]
        if mass_weight:
            modes = modes.make_mass_weighted()
        else:
            modes = modes.remove_mass_weighting()
            if mass_scale:
                extent /= np.sqrt(
                    np.dot(np.repeat(self.atomic_masses, 3), base_expansion[0][which]**2)
                        * UnitsData.convert("ElectronMass", "AtomicMassUnits")
                )
        if frequency_scale:
            freqs = np.abs(modes.freqs)
            extent *= np.sqrt(np.min(freqs) / freqs[which])
        if order is not None:
            base_expansion = ModeEmbedding(self.embedding, modes).get_cartesians_by_internals(order=order)
        if coordinate_expansion is not None:
            base_expansion = nput.tensor_reexpand(coordinate_expansion, base_expansion)
        return self.animate_coordinate(which,
                                       extent=extent, steps=steps,
                                       coordinate_expansion=base_expansion,
                                       **opts
                                       )


    def _format_xyz(self, which, nat, atoms, geom, float_format='10.3f'):
        xyz_elems = len(geom[0])
        template = "{atom} " + " ".join(f"{{xyz[{i}]:{float_format}}}" for i in range(xyz_elems))
        body = "\n".join(template.format(atom=atom, xyz=xyz) for atom, xyz in zip(atoms, geom))
        return f"{nat}\nstruct {which}\n{body}"

    def format_structs(self,
                      geoms,
                      format='xyz'
                    ):
        geom_data = np.reshape(np.asanyarray(geoms), (-1, len(self.atoms), 3))
        atoms = self.atoms
        if format == 'xyz':
            nat = len(atoms)
            return "\n".join(
                self._format_xyz(which + 1, nat, atoms, geom)
                for which, geom in enumerate(geom_data)
            )
        else:
            raise NotImplementedError(format)
    def _format_jmol_displacements_files(self, coords, expansion, units='Angstroms'):
        atoms = self.atoms
        nat = len(atoms)
        disps = expansion[0].reshape(-1, nat, 3)
        if units is not None:
            coords = coords * UnitsData.convert("BohrRadius", units)
            disps = disps / UnitsData.convert("BohrRadius", units)
        # raise Exception(coords.shape, disp[0].shape)
        return [
            self._format_xyz(
                i + 1,
                nat,
                atoms,
                np.concatenate([coords, np.zeros((nat, 1)), disp], axis=1)
            )
            for i,disp in enumerate(disps)
        ]

    def format_animation_file(self, which, format='xyz', extent=.35, steps=8, strip_embedding=True, units='Angstroms',
                              coordinate_expansion=None
                              ):
        if format == 'jmol':
            if coordinate_expansion is None:
                coordinate_expansion = self.get_cartesians_by_internals(1, strip_embedding=strip_embedding)
            coordinate_expansion = coordinate_expansion[0]
            return self._format_jmol_displacements_files(
                self.coords,
                [coordinate_expansion[(which,), :] * extent],
                units
            )[0]
        else:
            geoms = self.get_animation_geoms(which, extent=extent, steps=steps, strip_embedding=strip_embedding, units=units,
                                             coordinate_expansion=coordinate_expansion)
            return self.format_structs(geoms, format)

    def jsmol_viz(self, xyz=None, animate=False, vibrate=False, script=None):
        from McUtils.Jupyter import JSMol
        if xyz is None:
            xyz = self._format_xyz(0,
                                   len(self.atoms),
                                   self.atoms,
                                   self.coords * UnitsData.convert("BohrRadius", "Angstroms")
                                   )
        return JSMol.Applet(xyz, animate=animate, vibrate=vibrate, load_script=script)

    def jupyter_viz(self):
        from McUtils.Jupyter import MoleculeGraphics

        return MoleculeGraphics(self.atoms,
                                np.ndarray.view(self.coords.convert(CartesianCoordinates3D)),
                                bonds=self.bonds
                                )

    def to_widget(self):
        if self.display_jsmol:
            obj = self.plot(mode='jsmol', return_objects=False)
        else:
            obj = self.plot(backend='x3d', return_objects=False).figure.to_x3d()
        return obj

    default_display_mode = 'jsmol'
    def _ipython_display_(self):
        obj = self.plot(mode=self.default_display_mode, return_objects=False)
        return obj._ipython_display_()
        # return self.jupyter_viz()._ipython_display_()
    #endregion

    #region External Program Properties
    def _get_ob_attr(self, item):
        if self._mol is None:
            raise AttributeError("No pybel molecule")
        else:
            return getattr(self._mol, item)
    #endregion

class MolecoolException(Exception):
    pass



"""
The main package for working with core coordinate/molecule/wavefunction stuff.
Provides base classes and common/shared interfaces as well as a number of domain-specific applications.
The packages exposed include: `Molecools`, a handler for molecules (or really systems of atoms) with a
rich set of query-able properties; `DVR`, a package for arbitrary dimensional discrete variable representations
with the ability to generalize easily with full kinetic couplings; and `VPT2`, a misleadingly named package that
implements arbitrary order perturbation theory with flexible coordinate choice.
There are also a number of helper packages implementing things like a wavefunction interface, basis set representations,
and spectrum handling.
"""

import Psience.Molecools as Molecools
import Psience.Wavefun as Wavefun
import Psience.DVR as DVR
import Psience.VPT2 as VPT2
import Psience.Data as Data
import Psience.BasisReps as BasisReps
import Psience.AnalyticModels as AnalyticModels

# getting the full list of symbols explicitly in an __all__ variable
__all__ = [
    "Molecools",
    "Wavefun",
    "DVR",
    "VPT2",
    "Data",
    "BasisReps",
    "Spectra",
    "AnalyticModels",
    "Modes"
]


import numpy as np, math

from McUtils.Coordinerds import CartesianCoordinates1D, CartesianCoordinates2D, CartesianCoordinates3D
from McUtils.Scaffolding import Logger
from McUtils.Parallelizers import Parallelizer
from McUtils.Data import AtomData, UnitsData
import McUtils.Numputils as nput

from ..Modes import NormalModes

from .Coordinates import *
from .Evaluators import *
from .Solvers import *
from .Interpolation import *

__all__ = [
    "DGBGaussians"
]

class DGBCovarianceTransformations:

    def __init__(self, tfs, invs=None, inverse_sum=None):
        self.tfs = tfs
        if invs is None and self.tfs is not None:
            dets = np.linalg.det(tfs)
            if not np.allclose(dets, np.ones(len(dets))):
                raise ValueError("inverse must be supplied for non-unitary transforms")
            else:
                invs = tfs.transpose(0, 2, 1)
        self.invs = invs
        self.inverse_sum = inverse_sum

    def transform(self, derivative_operators):
        raise NotImplementedError("proper transformation support still to come")

    def inverse(self):
        return type(self)(self.invs, self.tfs)

    def combine(self, other:'DGBCovarianceTransformations'):
        if self.tfs is None and other.tfs is None:
            tfs = None
            invs = None
            inverse_sum = None
        elif self.tfs is None or other.tfs is None:
            raise NotImplementedError("if one set of covariances is specified, both must be")
        else:
            inverse_sum = np.linalg.inv(self.invs + other.invs)
            tfs = self.tfs + other.tfs
            invs = other.invs - other.invs @ inverse_sum @ other.invs

        return type(self)(tfs, invs, inverse_sum=inverse_sum)

class DGBGaussians:
    """
    A class to set up the actual N-dimensional Gaussians used in a DGB
    """

    def __init__(self, coords, alphas,
                 transformations=None, *,
                 momenta=None,
                 poly_coeffs=None,
                 kinetic_options=None,
                 logger=None,
                 parallelizer=None
                 ):
        self._pref = None
        self._S = None
        self._T = None

        if not isinstance(coords, DGBCoords):
            raise ValueError("need DGBCoords passed in")

        self.coords = coords
        if isinstance(alphas, (int, float, np.integer, np.floating)):
            alphas = [alphas] * self.coords.shape[0]
        self.alphas = np.asanyarray(alphas)


        if self.alphas.ndim == 1:
            self.alphas = np.broadcast_to(self.alphas[:, np.newaxis], self.coords.shape)
        self._transforms = self.canonicalize_transforms(self.alphas, transformations)

        if isinstance(momenta, (int, float, np.integer, np.floating)):
            momenta = [momenta] * self.coords.shape[0]
        if momenta is not None:
            momenta = np.asanyarray(momenta)
            if momenta.ndim == 1:
                momenta = np.broadcast_to(momenta[:, np.newaxis], self.alphas.shape)
        self.momenta = momenta

        self._overlap_data = None

        if poly_coeffs is not None:
            poly_coeffs = self.canonicalize_poly_coeffs(poly_coeffs, self.alphas)
        self._poly_coeffs = poly_coeffs

        if kinetic_options is None:
            kinetic_options = {}
        self.kinetic_options = kinetic_options

        self.parallelizer = Parallelizer.lookup(parallelizer)
        self.logger = Logger.lookup(logger)

    @property
    def overlap_data(self):
        if self._overlap_data is None:
            self._overlap_data = DGBEvaluator.get_overlap_gaussians(
                self.coords.centers,
                self.alphas,
                self.transformations,
                self.momenta,
                logger=self.logger,
                parallelizer=self.parallelizer
            )
        return self._overlap_data
    def get_S(self, return_prefactor=False):
        if self._poly_coeffs is not None:
            raise NotImplementedError("need to reintroduce polynomial support")
        with self.logger.block(tag="Evaluating S matrix"):
            return DGBEvaluator.evaluate_overlap(self.overlap_data, logger=self.logger, return_prefactor=return_prefactor)
    def get_T(self):
        if self._poly_coeffs is not None:
            raise NotImplementedError("need to reintroduce polynomial support")
        with self.logger.block(tag="Evaluating kinetic energy matrix"):
            return self.coords.kinetic_energy_evaluator.evaluate_ke(self.overlap_data, logger=self.logger, **self.kinetic_options)

    def optimize(self, optimizer_options, potential_function=None, logger=None, **opts):
        if logger is None:
            logger = self.logger
        if optimizer_options is True:
            optimizer_options = 'gram-schmidt'
        if isinstance(optimizer_options, (int, float, np.integer, np.floating)):
            optimizer_options = {'method':'gram-schmidt', 'norm_cutoff':optimizer_options}
        if isinstance(optimizer_options, str):
            optimizer_options = {'method':optimizer_options}

        # dispatch
        method = optimizer_options['method'].lower()
        opts = dict(optimizer_options, **opts)
        del opts['method']
        if method == 'svd':
            optimized_positions = self._optimize_svd(logger=logger, **opts)
        elif method == 'gram-schmidt':
            optimized_positions = self._optimize_gs(logger=logger, **opts)
        elif method == 'energy-cutoff':
            optimized_positions = self._prune_energy(potential_function, logger=logger,  **opts)
        elif method == 'decluster':
            optimized_positions = self._prune_dists(logger=logger, **opts)
        else:
            raise ValueError("don't know what to do with optimizer method '{}'".format(method))

        return self.take_gaussian_selection(optimized_positions)

    def take_gaussian_selection(self, full_good_pos):
        centers = self.coords[full_good_pos,]
        alphas = self.alphas[full_good_pos,]
        if self.transformations is not None:
            transformations = [t[full_good_pos,] for t in self.transformations]
        else:
            transformations = None

        if self._poly_coeffs is not None:
            _poly_coeffs = [t[full_good_pos,] for t in self._poly_coeffs]
        else:
            _poly_coeffs = None

        if self.momenta is not None:
            momenta = self.momenta[full_good_pos,]
        else:
            momenta = None

        new = type(self)(
            centers, alphas, transformations,
            poly_coeffs=_poly_coeffs, logger=self.logger,
            momenta=momenta,
            kinetic_options=self.kinetic_options
        )
        
        if self._overlap_data is not None:
            new._overlap_data = self.overlap_data.take_subselection(full_good_pos)
        if self._pref is not None:
            base = self._pref
            idx = np.ix_(full_good_pos, full_good_pos)
            if isinstance(base, tuple):
                new._pref = (base[0][idx], base[1][idx])
            else:
                new._pref = base[idx]
        if self._S is not None:
            base = self._S
            idx = np.ix_(full_good_pos, full_good_pos)
            if isinstance(base, tuple):
                new._S = (base[0][idx], base[1][idx])
            else:
                new._S = base[idx]
        if self._T is not None:
            base = self._T
            idx = np.ix_(full_good_pos, full_good_pos)
            if isinstance(base, tuple):
                new._T = (base[0][idx], base[1][idx])
            else:
                new._T = base[idx]

        return new

    @staticmethod
    def _calc_gs_norm(new, S, old):
        # calculates a very particular form of norm where (for speed purposes)
        #   1. new is the new _unnormalized_ orthogonal vector
        #   2. old is the row of the overlap matrix pulled to orthogonalize
        #   3. S is the subblock of the overlap matrix corresponding to everything up to the newest element
        #
        # This means we get to calculate a basic norm corresponding to the previous block by using everything
        # up to the final element of new and then we add on a correction knowing that the missing piece of S
        # is encoded in old

        vec = new[:-1]
        rows, cols = np.triu_indices(len(vec), k=1)
        off_diag = 2 * np.dot(S[rows, cols], vec[rows] * vec[cols])
        diag = np.dot(vec, vec) # taking advantage of the fact that the diagonal of S is 1
        base_norm = diag + off_diag

        # correction from old itself, knowing that the final element of new is 1
        correction_norm = 1 + 2 * np.dot(old[:-1], vec)

        return base_norm + correction_norm
    @staticmethod
    def _calc_gs_next(vec, ortho):
        new = np.zeros(len(vec))
        new[-1] = 1
        for prev in ortho:
            n = len(prev)
            new[:n] -= np.dot(vec[:n], prev) * prev
        return new

    _fast_outer = None
    @classmethod
    def _gs_off_diag_norms(cls, S, rows, cols, vec): # this is the slow step...
        if cls._fast_outer is None:
            import time
            # from McUtils.Misc import jit
            # @jit(nopython=True, cache=True, warn=True)
            def fast_mul(vec:np.ndarray, rows:np.ndarray, cols:np.ndarray):
                # wtf = time.time()
                mul = (vec[:, rows] * vec[:, cols])
                # elp = time.time() - wtf
                # print("--->", vec.shape, np.average(np.sum(np.abs(vec) > 1e-3, axis=1)), len(rows), f"{elp}s")
                return mul
            cls._fast_outer = fast_mul


        return 2 * np.matmul(
            S[rows, cols][np.newaxis, np.newaxis, :],
            cls._fast_outer(vec, rows, cols)[:, :, np.newaxis]
        ).reshape(-1)
    @staticmethod
    def _gs_diag_norms(vec):
        return np.matmul(
            vec[:, np.newaxis, :],
            vec[:, :, np.newaxis]
        ).reshape(-1)  # taking advantage of the fact that the diagonal of S is 1
    @staticmethod
    def _gs_extra_norm(old, vec):
        # correction from old itself, being a column of the overlap matrix, knowing that the final element of new is 1
        return 1 + 2 * np.matmul(
            old[:, np.newaxis, :], vec[:, :, np.newaxis]
        ).reshape(-1)
    # @staticmethod
    # def _gs_direct_norm(S, vec):
    #     print(S.shape, vec.shape)
    #     ugh = S[:len(vec), :len(vec)]
    #     Sv = np.matmul(ugh, vec).reshape(vec.shape)
    #     return np.matmul(Sv[:, np.newaxis, :], vec[:, :, np.newaxis]).reshape(-1)
    @classmethod
    def _gs_direct_norm(cls, S, vec):
        c1 = np.dot(vec, S)# I just assume dot is faster
        return np.matmul(
            c1[:, np.newaxis, :],
            vec[:, :, np.newaxis]
        ).flatten()
    @classmethod
    def _calc_gs_norm_block(cls, new, S, old):
        # generalizes the row-by-row form to calculate the norms for every vector at once so
        # we can figure out which ones to keep

        # This provides the contribution from (k-1)x(k-1)
        # of older terms
        vec = new[:, :-1]
        rows, cols = np.triu_indices(vec.shape[-1], k=1)

        # off_diag = cls._gs_off_diag_norms(S, rows, cols, vec)
        # diag = cls._gs_diag_norms(vec)
        # base_norm = diag + off_diag
        subs = np.eye(vec.shape[-1])
        subs[rows, cols] = S[rows, cols]
        subs[cols, rows] = S[rows, cols]
        base_norm = cls._gs_direct_norm(subs, vec)
        correction_norm = cls._gs_extra_norm(old, vec)

        return base_norm + correction_norm
    @staticmethod
    def _calc_gs_next_block(vecs, ortho):
        new = np.zeros(vecs.shape)
        new[:, -1] = 1
        for prev in ortho:
            n = len(ortho)
            new[:, :n] -= np.dot(vecs[:, :n], prev[:n])[:, np.newaxis] * prev[np.newaxis, :n]
        return new
    @staticmethod
    def _calc_gs_next_block_single(cur, overlaps, prev):
        cur = cur.copy()
        cur[:, -1] = 1
        n = len(prev)
        changes = np.dot(overlaps[:, :n], prev[:n])[:, np.newaxis] * prev[np.newaxis, :n]
        cur[:, :n] -= changes
        return changes, cur
    @classmethod
    def _optimize_gs_iter(cls, S, overlap_cutoff, norm_truncation_cutoff, logger=None):
        orthog_vecs = [np.array([1.])]
        norms = []  # debug
        mask = np.full(len(S), False, dtype=bool)
        mask[0] = True
        logger.log_print("pruning with unpivoted Gram-Schmidt norm > {n}", n=norm_truncation_cutoff)
        for i in range(1, len(S)):
            mask[i] = True
            vec = S[i, :i + 1][mask[:i + 1]]
            new = cls._calc_gs_next(vec, orthog_vecs)
            norm = cls._calc_gs_norm(new, S[np.ix_(mask[:i], mask[:i])], vec)
            norms.append(norm)
            if norm > overlap_cutoff:
                if norm < norm_truncation_cutoff:
                    norm = norm_truncation_cutoff
                orthog_vecs.append(new / np.sqrt(norm))
            else:
                mask[i] = False

        return np.where(mask)[0]

    @staticmethod
    def _pivot_vector(vec, i, pivot):
        new_pivot = vec[i + pivot]
        vec[i + 1:i + 1 + pivot] = vec[i:i + pivot]
        vec[i] = new_pivot
        return vec
    @classmethod
    def _pivot_matrix(cls, mat, i, pivot):
        pivot_row = cls._pivot_vector(mat[i + pivot, :].copy(), i, pivot)
        pivot_col = cls._pivot_vector(mat[:, i + pivot].copy(), i, pivot)
        pivot_pos = i + pivot

        # Unnecessarily complicated, maybe more efficient...
        mat[i + 1:pivot_pos + 1, :] = mat[i:pivot_pos, :]
        mat[:, i + 1:pivot_pos + 1] = mat[:, i:pivot_pos]
        mat[i, :] = pivot_row
        mat[:, i] = pivot_col
        mat[i, i+1:pivot_pos+1] = pivot_col[i+1:pivot_pos+1]
        mat[i+1:pivot_pos+1, i] = pivot_row[i+1:pivot_pos+1]

        return mat

    @classmethod
    def _optimize_gs_block(cls, S, overlap_cutoff, max_overlap_component,
                           # recalculate_cutoff_factor=0,
                           # chunk_size=5
                           logger=None
                           ):

        logger.log_print("pruning with pivoted Gram-Schmidt norm > {oc}", oc = overlap_cutoff)

        pivots = np.arange(len(S))
        # S_og = S
        S = S.copy()  # we'll use the upper triangle for norms and the lower triangle for storing vecs
        cols, rows = np.triu_indices_from(S, k=1)
        S[rows, cols] = 0 # remove lower triangle to redefine later
        if max_overlap_component < 1: # a fast pre-filter
            bad_pos = np.where(S[cols, rows] > max_overlap_component)
            if len(bad_pos[0]) > 0:
                # we contract S to reflect all of the things we've dropped
                kept = np.setdiff1d(np.arange(len(S)), rows[bad_pos[0]])
                pivots = pivots[kept]
                full_spec = kept
                S = S[np.ix_(full_spec, full_spec)]
                logger.log_print("dropped {n} Gaussians with overlaps greater than {mc}",
                                 n=len(np.unique(rows[bad_pos[0]])),
                                 mc=max_overlap_component)
        # all_norms = None
        for i in range(1, len(S)):

            all_vecs = S[i:, :i + 1]
            all_ovs = S[:i + 1, i:].T
            # all_news = cls._calc_gs_next_block(all_vecs, [S[:j+1] for j in range(i)])  # subtract off newest change
            changes, all_news = cls._calc_gs_next_block_single(all_vecs, all_ovs, S[i-1, :i])  # subtract off newest change
            all_norms = cls._calc_gs_norm_block(all_news, S, S[:i, i:].T)

            # rem = len(pivots) - i
            good_pos = np.where(all_norms > overlap_cutoff)
            # print("-->", len(good_pos[0]))
            if len(good_pos) == 0 or len(good_pos[0]) == 0:
                # we know everything else can be eliminated
                pivots = pivots[:i]
                break
            else:
                # we contract S to reflect all of the things we've dropped
                kept = good_pos[0]
                pivots = np.concatenate([pivots[:i], pivots[i + kept]])
                if len(kept) == 1:
                    break

                full_spec = np.concatenate([np.arange(i), i + kept])
                S = S[np.ix_(full_spec, full_spec)]
                # S_og = S_og[np.ix_(full_spec, full_spec)]
                all_news = all_news[kept,]
                all_norms = all_norms[kept,]

            S[i:, :i] = all_news[:, :i]  # update without normalizing for now
            pivot_pos = np.argmax(all_norms)  # find the position with the largest remaining norm
            new_vec = all_news[pivot_pos] / np.sqrt(all_norms[pivot_pos])  # normalize just this vector
            # print(all_norms[pivot_pos])

            # Rearrange the pivots now that we have the new ordering
            pivots = cls._pivot_vector(pivots, i, pivot_pos)
            # all_norms = cls._pivot_vector(all_norms, i, pivot_pos)

            # rearrange S to reflect this new ordering
            S = cls._pivot_matrix(S, i, pivot_pos)
            # S_og = cls._pivot_matrix(S_og, i, pivot_pos)

            # now insert the normalized vector
            S[i, :i+1] = new_vec
            # print(S[:i+1, :i+1])
            # print(S_og[:i+1, :i+1])
            # if i > 1:
            #     raise Exception(...)
        return np.sort(pivots)

    # def _S_generator(self,
    #                  chunk_start, chunk_size
    #                  ):
    #
    #     DGBEvaluator.evaluate_overlap(self.overlap_data, logger=self.logger)
    #     self.get_S()
    gs_optimization_overlap_cutoff=1e-3
    def _optimize_gs(self, *, S=None,
                     norm_cutoff=None,
                     norm_truncation_cutoff=0,
                     max_overlap_cutoff=1,
                     allow_pivoting=True,
                     chunk_size=None,
                     logger=None
                     ):

        with logger.block(tag='optimizing using Gram-Schmidt'):
            if norm_cutoff is None:
                norm_cutoff = self.gs_optimization_overlap_cutoff
            if S is None:
                if chunk_size is None:
                    S=self.S
                    if isinstance(S, tuple): # phases introduced
                        S = 1/2*(S[0] + S[1])
                else:
                    raise NotImplementedError('getting chunking working is hard...')
                    S=self.S

            if allow_pivoting:
                pivots = self._optimize_gs_block(S, norm_cutoff, max_overlap_cutoff, logger=logger)
            else:
                pivots = self._optimize_gs_iter(S, norm_cutoff, norm_truncation_cutoff, logger=logger)
            logger.log_print("pruned {nD} Gaussians",  nD=len(S)-len(pivots))

        return pivots

    def _optimize_svd(self, *, S=None,
                      min_value=1e-12,
                      num_vectors=None,
                      contrib_cutoff=1e-3,
                      logger=None
                      ):
        if S is None:
            S = self.S

        with logger.block(tag='optimizing by taking subspaces of the overlap matrix'):
            sig, evecs = np.linalg.eigh(S)
            # for i in range(5):
            #     print("wtf", i, len(np.unique(np.where(np.abs(evecs[:, -(i+1):]) > contrib_cutoff)[0])), sig[-(i+1)],
            #           np.min(evecs[:, -(i+1)]), np.max(evecs[:, -(i+1)]), np.std(evecs[:, -(i+1)]))
            if num_vectors:
                good_loc = slice(max(len(sig) - num_vectors, 0), len(sig))
            else:
                # self.logger.log_print("most important center threshold: {t}", t=min_singular_value)
                good_loc = np.where(sig > min_value)[0]
            full_good_pos = np.unique(np.where(np.abs(evecs[:, good_loc]) > contrib_cutoff)[0])
            logger.log_print("pruned {nD} Gaussians",  nD=len(S)-len(full_good_pos))

        return full_good_pos

    default_energy_cutoff = 1600 / UnitsData.hartrees_to_wavenumbers
    def _prune_energy(self, pot, *, cutoff=None, probabilities=None, logger=None, potential_values=None):
        if potential_values is None:
            potential_values = pot(self.coords.centers)
        if probabilities is not None:
            with logger.block(tag="pruning Gaussians with energies by distribution {d} cm-1",
                              d=[
                                  [np.round(c * UnitsData.hartrees_to_wavenumbers, 3), 1 if p is None else p]
                                  for c, p in probabilities
                              ]):
                pots = np.sort(potential_values)
                cur_pivot = 0
                pivots = []
                for next_e, next_prob in probabilities:
                    pivot_pos = np.searchsorted(pots[cur_pivot:], next_e)+1 # position past where the energy would occur
                    if next_prob is None:
                        new_pivs = np.arange(cur_pivot, pivot_pos)
                    else:
                        if next_prob < 1:
                            probs = np.random.uniform(0, 1, size=pivot_pos)
                            sel = np.where(probs < next_prob)
                        else: # take this many points from that region
                            if pivot_pos >= next_prob:
                                sel = [
                                    np.sort(np.random.choice(np.arange(pivot_pos-1), next_prob, replace=False))
                                ]
                            else:
                                sel = [np.arange(pivot_pos-1)]
                        if len(sel) > 0 and len(sel[0]) > 0:
                            new_pivs = cur_pivot + sel[0]
                    cur_pivot = pivot_pos
                    pivots.append(new_pivs)
                full_good_pos = np.concatenate(pivots)
                logger.log_print("pruned {nD} Gaussians", nD=len(potential_values) - len(full_good_pos))
        else:
            if cutoff is not None:
                cutoff = self.default_energy_cutoff
            with logger.block(tag="pruning Gaussians with energies larger than {ec} cm-1",
                              ec=np.round(cutoff * UnitsData.hartrees_to_wavenumbers, 3)
                              ):
                full_good_pos = np.where(potential_values < cutoff)[0]
                logger.log_print("pruned {nD} Gaussians", nD=len(potential_values) - len(full_good_pos))
        return full_good_pos

    def _prune_dists(self, *,  cluster_radius, logger=None):
        with logger.block(tag=logger.log_print('declustering data with a radius of {r}', r=cluster_radius)):
            points = self.coords.centers
            pivots = np.arange(len(points))
            dec_pts = points.reshape(points.shape[0], -1)
            for i in range(len(points)):
                cur_pos = pivots[i]
                dists = np.linalg.norm(
                    dec_pts[cur_pos][np.newaxis, :] - dec_pts[pivots[i + 1:], :],
                    axis=1
                )
                good_pos = np.where(dists > cluster_radius)
                if len(good_pos) == 0 or len(good_pos[0]) == 0:
                    break
                pivots = np.concatenate([pivots[:i + 1], pivots[i + 1:][good_pos]])
            logger.log_print("pruned {nD} Gaussians", nD=len(self.coords.centers) - len(pivots))
        return pivots

    @classmethod
    def construct(cls,
                  coords,
                  alphas,
                  *,
                  potential_expansion=None,
                  potential_function=None,
                  transformations=None,
                  masses=None,
                  atoms=None,
                  modes=None,
                  kinetic_options=None,
                  internals=None,
                  coordinate_selection=None,
                  cartesians=None,
                  gmat_function=None,
                  momenta=None,
                  poly_coeffs=None,
                  logger=None,
                  pairwise_potential_functions=None,
                  parallelizer=None
                  ):
        if potential_expansion is not None:
            if potential_function is not None:
                raise ValueError("can't have both `potential_function` and `potential_expansion`")
            if isinstance(potential_expansion, dict):
                expansion_opts = potential_expansion.copy()
                del expansion_opts['values']
                potential_expansion = potential_expansion['values']
                interp_coords = expansion_opts.get('centers', coords)
                if 'centers' in expansion_opts:
                    del expansion_opts['centers']
            else:
                interp_coords = coords
                expansion_opts = {}
            potential_function = DGBGenericInterpolator(
                interp_coords,
                potential_expansion,
                pairwise_potential_functions=pairwise_potential_functions,
                **expansion_opts
            )

        if not isinstance(coords, DGBCoords):
            if modes is not None:
                if isinstance(modes, str):
                    if modes == 'normal':
                        modes = cls.get_normal_modes(
                            coords,
                            potential_function,
                            masses=masses,
                            atoms=atoms,
                            internals=internals,
                            gmat_function=gmat_function
                        )

                    else:
                        raise ValueError("unknown mode spec {}".format(modes))

                if coords.ndim == 3: # Cartesian-like
                    # coords_old = coords

                    coords = DGBWatsonModes.from_cartesians(
                        coords,
                        modes,
                        masses=DGBCartesians.resolve_masses(
                            coords,
                            masses=masses,
                            atoms=atoms
                        )
                    )

                    if momenta is not None: # defined as d/dx, even though classically given by m dx/dt
                        momenta = [np.zeros(momenta.shape[0]), momenta.reshape((momenta.shape[0], -1))]
                        momenta = DGBWatsonModes.embed_derivs(momenta, modes)[1]

                else:
                    if internals is not None:
                        raise NotImplementedError("internal modes not implemented")
                    else:
                        raise NotImplementedError("transformation not fully implemented")
                        # Is this right???
                        coords = DGBWatsonModes(
                            modes.matrix @ coords,
                            modes=modes
                        )

                gmat_function = coords.gmatrix

            elif internals is not None:
                raise NotImplementedError("internal coordinate mapping not supported yet...")
            else:
                coords = DGBCartesians.from_cartesians(coords, masses=masses, atoms=atoms)
                if cartesians is not None:
                    coords = coords[:, :, cartesians]

            if coordinate_selection is not None:
                coords = coords.take_indices(coordinate_selection)

            potential_function = coords.embed_function(potential_function)

        if transformations is not None:
            if gmat_function is None:
                gmat_function = coords.gmatrix
            if isinstance(transformations, dict):
                tf_opts = dict({'gmat_function':gmat_function}, **transformations)
                transformations = tf_opts['method']
                del tf_opts['method']
            else:
                tf_opts = {'gmat_function':gmat_function}
            if isinstance(transformations, str):
                if transformations == 'rpath':
                    transformations = cls.get_reaction_path_transformations(
                        coords.coords,
                        potential_function,
                        # masses=masses,
                        # atoms=atoms,
                        # internals=internals,
                        **tf_opts#gmat_function=gmat_function
                    )
                elif transformations == 'diag':
                    transformations = cls.get_hessian_diagonalizing_transformations(
                        coords.coords,
                        potential_function,
                        # masses=masses,
                        # atoms=atoms,
                        # internals=internals,
                        **tf_opts
                    )
                else:
                    raise ValueError("unknown transformation spec {}".format(modes))
            transformations = cls.canonicalize_transforms(coords.centers, transformations)

            if momenta is not None:
                momenta = DGBEvaluator.get_momentum_vectors(momenta, transformations)

        if isinstance(alphas, (str, dict)):
            if isinstance(alphas, str) and alphas == 'auto':
                if modes is not None:
                    alphas = 'virial'
                else:
                    alphas = 'masses'
            alphas = cls.dispatch_get_alphas(
                alphas,
                coords,
                masses=masses,
                gmat_function=coords.gmatrix if gmat_function is None else gmat_function,
                potential_function=potential_function,
                transformations=transformations
            )

        shp = coords.shape
        if isinstance(alphas, (int, float, np.integer, np.floating)):
            alphas = np.full(shp, alphas)
        else:
            alphas = np.asanyarray(alphas)
            alphas = np.broadcast_to(alphas, shp)

        return cls(coords, alphas, transformations,
                   poly_coeffs=poly_coeffs, logger=logger,
                   momenta=momenta,
                   kinetic_options=kinetic_options
                   ), potential_function
    @classmethod
    def get_normal_modes(cls,
                         coords,
                         potential_function,
                         masses=None,
                         atoms=None,
                         internals=None,
                         gmat_function=None,
                         reference_structure=None,
                         stationary_point_norm=1e-2
                         ):
        if internals is not None:
            raise ValueError("internal coordinate support still to come")
        if reference_structure is None:
            f_data = potential_function(coords, deriv_order=2)
            if isinstance(f_data, np.ndarray) and f_data.ndim == 3:
                hess = f_data
                grad = potential_function(coords, deriv_order=1)
                pot = potential_function(coords, deriv_order=1)
            else:
                pot, grad, hess = f_data
            sorting = np.argsort(pot)
            coords = coords[sorting,]
            grad = grad[sorting,]
            hess = hess[sorting,]
            grad_norms = np.linalg.norm(grad, axis=-1)
            stationary_structures = np.where(grad_norms < stationary_point_norm)
            if len(stationary_structures) == 0 or len(stationary_structures[0]) == 0:
                raise ValueError("no stationary state found?")
            stationary_structures = stationary_structures[0]
            reference_structure = coords[stationary_structures[0]]
            f_matrix = hess[stationary_structures[0]]
        else:
            f_matrix = potential_function(reference_structure, deriv_order=2)
            if not isinstance(f_matrix, np.ndarray):
                f_matrix = f_matrix[2]

        if gmat_function is not None:
            mass_spec = gmat_function(reference_structure)
        else:
            mass_spec = DGBCartesians.resolve_masses(coords, masses=masses, atoms=atoms)

        if coords.shape[-1] == 3:
            basis = CartesianCoordinates3D
        elif coords.shape[-1] == 2:
            basis = CartesianCoordinates2D
        elif coords.shape[-1] == 1:
            basis = CartesianCoordinates1D
        else:
            raise ValueError("higher than 3D Cartesians?")

        return NormalModes.from_fg(
            basis,
            f_matrix,
            mass_spec,
            origin=reference_structure,
            dimensionless=False
        )

    @classmethod
    def get_reaction_path_transformations(
            cls,
            coords,
            potential_function,
            gmat_function,
            stationary_point_norm=1e-4,
            sort_alphas=True
    ):
        f_data = potential_function(coords, deriv_order=2)
        if isinstance(f_data, np.ndarray) and f_data.ndim == 3:
            hess = f_data
            grad = potential_function(coords, deriv_order=1)
            pot = potential_function(coords, deriv_order=1)
        else:
            pot, grad, hess = f_data

        hess_og = hess
        gmats = gmat_function(coords)
        # g12 = sp.linalg.fractional_matrix_power(gmats, 1 / 2)
        # gi12 = sp.linalg.fractional_matrix_power(gmats, -1 / 2)

        gvals, gvecs = np.linalg.eigh(gmats)
        if np.any((gvals <= 0).flatten()):
            raise ValueError("bad G-matrix?")
        g12_diags = np.zeros(gvecs.shape)
        diag_ings = (slice(None),) + np.diag_indices_from(gvecs[0])
        g12_diags[diag_ings] = np.sqrt(gvals)
        g12 = gvecs @ g12_diags @ gvecs.transpose(0, 2, 1)

        grad = np.reshape(grad[:, np.newaxis, :] @ g12, grad.shape) # mass-weight
        hess = g12 @ hess @ g12

        grad_norms = np.linalg.norm(grad, axis=-1)
        non_stationary = np.where(grad_norms >= stationary_point_norm)

        # in this case I'm _only_ projecting out the gradient for whatever that's worth...
        # which...probably has some rotation/translation component which
        # is why I don't get clean zero eigenvectors...
        rp_mode = -grad[non_stationary] / grad_norms[non_stationary][:, np.newaxis]
        proj = np.eye(rp_mode.shape[-1])[np.newaxis] - nput.vec_outer(rp_mode, rp_mode, axes=[1, 1])
        hess[non_stationary] = proj@hess[non_stationary]@proj

        freqs, g12_tfs = np.linalg.eigh(hess) # replace zero tf with mass-weighted rp_mode
        # raise Exception(np.sqrt(np.abs(freqs[7:15])) * 219475)

        # g12_diags[diag_ings] = 1 / np.sqrt(gvals)
        # g12_inv = gvecs @ g12_diags @ gvecs.transpose(0, 2, 1)

        tfs = g12 @ g12_tfs
        tfs[non_stationary, :, 0] = rp_mode

        # now we calculate the unweighted hessian diagonals to sort by
        if sort_alphas:
            tf_hess = tfs.transpose(0, 2, 1) @ hess_og @ tfs
            diags = np.argsort(np.abs(np.diagonal(tf_hess, axis1=1, axis2=2)), axis=1)

            tfs = tfs[
                np.arange(tfs.shape[0])[:, np.newaxis, np.newaxis],
                np.arange(tfs.shape[1])[np.newaxis, :, np.newaxis],
                diags[:, np.newaxis, :]
            ]

        inv = np.linalg.inv(tfs) #g12_tfs.transpose(0, 2, 1) @ g12_inv

        # return inv.transpose(0, 2, 1), tfs.transpose(0, 2, 1)
        return tfs, inv

    @classmethod
    def get_hessian_diagonalizing_transformations(
            cls,
            coords,
            potential_function,
            gmat_function
    ):
        f_data = potential_function(coords, deriv_order=2)
        if isinstance(f_data, np.ndarray) and f_data.ndim == 3:
            hess = f_data
            grad = potential_function(coords, deriv_order=1)
            pot = potential_function(coords, deriv_order=1)
        else:
            pot, grad, hess = f_data

        hess_og = hess
        gmats = gmat_function(coords)
        # g12 = sp.linalg.fractional_matrix_power(gmats, 1 / 2)
        # gi12 = sp.linalg.fractional_matrix_power(gmats, -1 / 2)

        gvals, gvecs = np.linalg.eigh(gmats)
        if np.any((gvals <= 0).flatten()):
            raise ValueError("bad G-matrix?")
        g12_diags = np.zeros(gvecs.shape)
        diag_ings = (slice(None),) + np.diag_indices_from(gvecs[0])
        g12_diags[diag_ings] = np.sqrt(gvals)
        g12 = gvecs @ g12_diags @ gvecs.transpose(0, 2, 1)

        hess = g12 @ hess @ g12

        _, g12_tfs = np.linalg.eigh(hess)

        tfs = g12 @ g12_tfs
        g12_diags[diag_ings] = 1 / np.sqrt(gvals)
        g12_inv = gvecs @ g12_diags @ gvecs.transpose(0, 2, 1)
        inv = g12_tfs.transpose(0, 2, 1) @ g12_inv

        # return tfs, inv
        return inv.transpose(0, 2, 1), tfs.transpose(0, 2, 1)

    @staticmethod
    def _filter_alpha_method_keys(method, opts, necessary_keys, optional_keys):
        new_opts = {}
        for k in necessary_keys:
            if k not in opts:
                raise ValueError(
                    "alpha inference method {} needs keys {}, missing {}".format(method, necessary_keys, k)
                )
            new_opts[k] = opts[k]
        for k in optional_keys:
            if k in opts: new_opts[k] = opts[k]
        return new_opts
    @classmethod
    def dispatch_get_alphas(self, alphas, centers, **extra_opts):
        if isinstance(alphas, str):
            alphas = {'method':alphas}
        opts = dict(extra_opts, **alphas)
        del opts['method']
        method = alphas['method']
        if method == 'virial':
            return self.get_virial_alphas(
                centers,
                **self._filter_alpha_method_keys(
                    'virial',
                    opts,
                    ['potential_function', 'gmat_function', 'transformations'],
                    ['scaling']
                )
            )
        elif method == 'masses':
            return self.get_mass_alphas(
                centers,
                **self._filter_alpha_method_keys(
                    'masses',
                    opts,
                    ['masses'],
                    ['scaling']
                )
            )
        # elif method == 'reaction_path':
        #     return self.get_virial_alphas(
        #         centers,
        #         **self._filter_alpha_method_keys(
        #             'virial',
        #             opts,
        #             ['potential_function', 'masses'],
        #             []
        #         )
        #     )
        elif method == 'min_dist':
            return self.get_min_distance_alphas(
                centers,
                **self._filter_alpha_method_keys(
                    'min_dist',
                    opts,
                    ['masses'],
                    ['scaling']
                )
            )
        else:
            raise ValueError("unknown method for getting alphas {}".format(alphas['method']))

    @classmethod
    def get_mass_alphas(cls, centers, *, masses, scaling=10, use_mean=False):
        h_mass =  AtomData["H", "Mass"] * UnitsData.convert("AtomicMassUnits", "AtomicUnitOfMass")
        a = scaling * np.asanyarray(masses) / h_mass
        nats = len(masses)
        ndim = centers.shape[-1] // nats
        a = np.broadcast_to(a[:, np.newaxis], (nats, ndim)).flatten()
        a = np.broadcast_to(a[np.newaxis], centers.shape)
        return a

    @classmethod
    def get_min_distance_alphas(cls, masses, centers, scaling=1/4, use_mean=False):
        # if clustering_radius is None:
        #     clustering_radius = 1
        # np.sqrt(masses / min_dist)
        distances = np.linalg.norm(centers[:, np.newaxis, :] - centers[np.newaxis, :, :], axis=-1)
        # mean_dist = np.average(distances[distances > 1e-8], axis=None)
        distances[distances < 1e-8] = np.max(distances) # exclude zeros
        closest = np.min(distances, axis=1)
        # too hard to compute convex hull for now...so we treat the exterior
        # the same as the interior
        a = scaling * np.sqrt(masses[np.newaxis, :] / closest[:, np.newaxis])
        if use_mean:
            a = np.mean(a)
        # print("???", a[:5])
        return a

    # TODO: add method to rotate to point along bonds
    #       not sure exactly how to define the rest of the coordinate system
    #       but in principle it could be isotropic...
    @classmethod
    def get_virial_alphas(cls,
                          coords,
                          *,
                          potential_function,
                          gmat_function,
                          transformations,
                          scaling=1/2
                          ):
        # we're just computing local frequencies along the coordinates defined by
        # applying the linear transformations given in "transformations" to the supplied coords

        #TODO: avoid recomputing this for like the 15th time
        if isinstance(coords, DGBCoords):
            # raise Exception(coords, coords.centers.shape)
            coords = coords.centers
        f_data = potential_function(coords, deriv_order=2)
        if isinstance(f_data, np.ndarray) and f_data.ndim == 3:
            hess = f_data
            # grad = potential_function(coords, deriv_order=1)
            # pot = potential_function(coords, deriv_order=1)
        else:
            pot, grad, hess = f_data

        gmats = gmat_function(coords)

        # raise Exception(gmats.shape, hess.shape)
        if transformations is not None:
            inv, transformations = transformations
            hess = nput.vec_tensordot(
                nput.vec_tensordot(
                    hess,
                    transformations,
                    axes=[1, 2],
                    shared=1
                ),
                transformations,
                axes=[1, 2],
                shared=1
            )

            gmats = nput.vec_tensordot(
                nput.vec_tensordot(
                    gmats,
                    transformations,
                    axes=[1, 2],
                    shared=1
                ),
                transformations,
                axes=[1, 2],
                shared=1
            )

        alphas = scaling * np.sqrt(np.abs(np.diagonal(hess, axis1=1, axis2=2) / np.diagonal(gmats, axis1=1, axis2=2)))
        # print(hess)
        # print(
        #     np.sqrt(np.abs(np.diagonal(hess, axis1=1, axis2=2) / np.diagonal(gmats, axis1=1, axis2=2))) * 219475
        # )
        # print(
        #     np.sum(np.diagonal(hess, axis1=1, axis2=2) * 219475, axis=1)
        # )
        #[ 9426.1752331  10317.82470038  9108.73436512  9714.64698828  8579.37664985  9764.6991486   9764.69922505  9197.73091615  9197.73093045 11040.69284146  8823.36978916  9911.15518915  7911.13613538 10086.93927288 10086.93926775  9071.32381428  9071.32391714 11374.03054902  8618.17129632  9967.37235458  7484.76311907 10195.248758   10195.24883265  8970.84073996  8970.84093599 11209.80670989  8524.28409883  9866.77241186  7325.03365781  9995.46064476  9995.46077198  8839.0596967   8839.05989418 10599.3940994   8556.00072571  9632.81612374  7438.73297697  9579.709646    9579.709779    8683.59676335  8683.59680886]
        #[ 9426.1752331  10317.97352922  9342.83704949 10024.66092923  8579.25827834  9748.71327209  9748.71335084  9210.28053516  9210.28054868 11040.48757709  9044.73381151 10280.23127104  7910.82531061 10003.62991937 10003.62993513  9078.13129445  9078.1314092  11373.51783935  8846.43367466 10397.23539095  7484.34647915 10082.52862084 10082.52870943  8972.7922302   8972.79246371 11210.08657008  8784.25964707 10318.11304479  7324.5703684   9957.98059176  9957.98072257  8823.24924889  8823.24949849 10602.16863697  8879.27410661  9971.67663237  7438.26430876  9726.22755942  9726.22773469  8638.84650038  8638.84655498]

        return alphas

    @staticmethod
    def _get_hermite_poly(coeff_dict, alphas):
        ndim = alphas.shape[-1]
        # raise Exception(
        #     np.flip(np.asarray(sp.special.hermite(coeff_dict.get(0, 0), monic=False))) *
        #         np.sqrt( (2 * alphas[0]) ** np.arange(coeff_dict.get(0, 0) + 1) )
        # )
        parts = [
            np.flip(np.asarray(sp.special.hermite(coeff_dict.get(k, 0), monic=False)))
            * np.sqrt(
                (2 * alphas[k] ) ** np.arange(coeff_dict.get(k, 0)+1)
                / ( 2**(coeff_dict.get(k, 0)) * math.factorial(coeff_dict.get(k, 0)) )
            )
            for k in range(ndim)
        ]
        coeffs = parts[0]
        for c in parts[1:]:
            coeffs = np.multiply.outer(coeffs, c)
        return coeffs
    @classmethod
    def canonicalize_poly_coeffs(cls, coeffs, alphas):
        if coeffs is None:
            return None

        poly_coeffs = [
            cls._get_hermite_poly(c, a) if isinstance(c, dict) else c
            for c, a in zip(coeffs, alphas)
        ]

        return poly_coeffs

    @property
    def transformations(self):
        return self._transforms
    @transformations.setter
    def transformations(self, tf):
        self._transforms = self.canonicalize_transforms(self.alphas, tf)
    @classmethod
    def canonicalize_transforms(self, coords, tfs):
        npts = coords.shape[0]
        ndim = coords.shape[1]
        if tfs is None:
            return tfs
        if isinstance(tfs, np.ndarray):
            dets = np.linalg.det(tfs)
            if not np.allclose(dets, np.ones(len(dets))):
                raise ValueError("inverse must be supplied for non-unitary transforms")
            invs = tfs.transpose((0, 2, 1))
            tfs = (tfs, invs)
        if len(tfs) != 2:
            raise ValueError("need both transformations and inverses")

        tfs, invs = tfs
        if tfs.shape[0] != npts:
            if tfs.shape[0] == 1:
                tfs = np.broadcast_to(tfs, (npts,) + tfs.shape[1:])
            else:
                raise ValueError("wrong number of transformations supplied")
        if invs.shape[0] != npts:
            if invs.shape[0] == 1:
                invs = np.broadcast_to(invs, (npts,) + invs.shape[1:])
            else:
                raise ValueError("wrong number of inverses supplied")
        if tfs.shape[2] != ndim:
            raise ValueError("transformations have wrong dimension got {}[2], expected {}".format(
                tfs.shape,
                ndim
            ))
        if invs.shape[1] != ndim:
            raise ValueError("inverses have wrong dimension got {}[1], expected {}".format(
                invs.shape,
                ndim
            ))

        return (tfs, invs)

    @property
    def prefactor(self):
        if self._pref is None:
            self._pref, self._S = self.get_S(return_prefactor=True)
        return self._pref
    @property
    def S(self):
        if self._S is None:
            self._pref, self._S = self.get_S(return_prefactor=True)
        return self._S
    @S.setter
    def S(self, smat):
        self._S = smat
    @property
    def T(self):
        if self._T is None:
            self._T = self.get_T()
        return self._T
    @T.setter
    def T(self, tmat):
        self._T = tmat

    bad_alpha_limit = 1e-15
    bad_scaling_limit = 1e-3
    def marginalize_out(self, indices, *, bad_alpha_limit=None, bad_scaling_limit=None):
        if bad_scaling_limit is None:
            bad_scaling_limit = self.bad_scaling_limit
        if bad_alpha_limit is None:
            bad_alpha_limit = self.bad_alpha_limit

        subcoords = self.coords.drop_indices(indices)

        ndim = self.coords.shape[-1]
        remaining = np.setdiff1d(np.arange(ndim), indices)

        if self.transformations is not None:
            full_covs = DGBEvaluator.get_inverse_covariances(
                1 / (4*self.alphas), # hacky but the inverse is expected to have 2a on the diagonal
                self.transformations
            )
            full_sel = np.arange(self.alphas.shape[0])

            dropcovs = full_covs[np.ix_(full_sel, indices, indices)]
            dropdets = np.linalg.det(dropcovs)
            zp = np.abs(dropdets) < bad_scaling_limit #TODO: fix this in case someone is in the numerically unstable limit...
            dropdets[zp] = 1
            scaling = 1 / dropdets
            scaling[zp] = bad_scaling_limit
            scaling *= np.power(np.pi, len(indices) / 4)

            subcovs = full_covs[np.ix_(full_sel, remaining, remaining)]
            inv_alphas, tfs = np.linalg.eigh(subcovs)
            inv_alphas[inv_alphas < bad_alpha_limit] = bad_alpha_limit

            alphas = 1 / (2*inv_alphas) # we have 1/2a
            tfs = (tfs, tfs.transpose(0, 2, 1))

            momenta = self.momenta
            if momenta is not None:
                # we use the fact that the momenta are along the same axes as the old alphas to get
                scaled_moms = self.alphas * self.momenta
                # now we have to deal with a S_A @ T @ p type term by noting that we can turn T into its
                # [indices x len(alphas)] subblock
                cov_mom = self.transformations[0][:, remaining, :] @ scaled_moms[:, :, np.newaxis]
                inv_cov = DGBEvaluator.get_inverse_covariances(alphas, tfs)
                momenta = np.reshape(inv_cov @ cov_mom, alphas.shape)

        else:
            scaling = np.power(2 * np.pi, len(indices) / 4) / np.power(np.prod(self.alphas[:, indices], axis=-1), 1 / 4)
            alphas = self.alphas[:, remaining]
            tfs = None
            momenta = self.momenta
            if momenta is not None:
                momenta = momenta[:, remaining]

        return scaling, type(self)(
            subcoords,
            alphas=alphas,
            transformations=tfs,
            momenta = momenta
        )

    def as_cartesians(self):
        if isinstance(self.coords, DGBCartesians):
            return self

        carts, (transformation, inverse) = self.coords.as_cartesians() # tfs is ncarts x nmodes
        if self.transformations is not None:
            tf, inv = self.transformations
            tfs = transformation[np.newaxis] @ tf
            invs = inv @ inverse[np.newaxis]
        else:
            tfs = np.broadcast_to(transformation[np.newaxis], (carts.shape[0],) + transformation.shape)
            invs = np.broadcast_to(inverse[np.newaxis], (carts.shape[0],) + inverse.shape)
        transformations = (tfs, invs)

        # num_extra = carts.shape[1] - self.alphas.shape[1]
        # alphas = np.concatenate(
        #     [self.alphas, np.zeros((carts.shape[0], num_extra))],
        #     axis=1
        # )
        alphas = self.alphas
        momenta = self.momenta
        return type(self)(
            carts,
            alphas,
            transformations=transformations,
            momenta=momenta
        )

    def plot_centers(self,
                     figure=None,
                     xyz_sel=None,
                     **plot_styles
                     ):
        import McUtils.Plots as plt

        if isinstance(self.coords, DGBCartesians):
            coords = self.coords.coords
            if xyz_sel is None and coords.shape[-1] > 2:
                raise ValueError("need an `xyz_sel` to be plottable")
            coords = coords[:, :, xyz_sel]
            if coords.shape[-1] > 2:
                raise ValueError("can't plot centers for more than 2D data...?")
            for atom in range(coords.shape[1]):
                if coords.shape[-1] == 1:
                    figure = plt.ScatterPlot(
                        coords[:, atom, 0],
                        np.zeros(len(coords)),
                        figure=figure,
                        **plot_styles
                    )
                else:
                    figure = plt.ScatterPlot(
                        coords[:, atom, 0],
                        coords[:, atom, 1],
                        figure=figure,
                        **plot_styles
                    )

        else:
            coords = self.coords.centers
            if coords.shape[-1] > 2:
                raise ValueError("can't plot centers for more than 2D data...?")

            if coords.shape[-1] == 1:
                figure = plt.ScatterPlot(
                    coords[:, 0],
                    np.zeros(len(coords)),
                    figure=figure,
                    **plot_styles
                )
            else:
                figure = plt.ScatterPlot(
                    coords[:, 0],
                    coords[:, 1],
                    figure=figure,
                    **plot_styles
                )

        return figure